import logging
import pytest
import csle_common.constants.constants as constants
from unittest.mock import MagicMock, patch
from csle_common.dao.emulation_config.emulation_env_config import EmulationEnvConfig
from csle_common.controllers.host_controller import HostController


class TestHostControllerSuite:
    """
    Test suite for host controller
    """

    @pytest.fixture(autouse=True)
    def emulation_env_config_setup(self) -> None:
        """
        Set up emulation environment configuration

        :return: None
        """
        # setup container configurations
        emulation_env_config = MagicMock(spec=EmulationEnvConfig)
        container1 = MagicMock()
        container1.physical_host_ip = "192.168.1.10"
        container1.docker_gw_bridge_ip = "172.17.0.1"
        container1.name = "spark-container-1"
        container1.get_ips.return_value = ["192.168.1.2"]
        container1.get_full_name.return_value = "container_name"

        container2 = MagicMock()
        container2.physical_host_ip = "192.168.1.11"
        container2.docker_gw_bridge_ip = "172.17.0.2"
        container2.name = "spark-container-2"
        containers_config = MagicMock()
        emulation_env_config.containers_config = containers_config
        emulation_env_config.containers_config.containers = [container1, container2]
        # setup kafka container
        kafka_container = MagicMock()
        kafka_container.physical_host_ip = "192.168.1.10"
        kafka_container.docker_gw_bridge_ip = "172.17.0.3"
        kafka_config = MagicMock()
        emulation_env_config.kafka_config = kafka_config
        emulation_env_config.kafka_config.container = kafka_container
        emulation_env_config.kafka_config.container.get_ips.return_value = ["192.168.1.2"]
        emulation_env_config.kafka_config.kafka_port = 9092
        emulation_env_config.kafka_config.time_step_len_seconds = 10
        emulation_env_config.kafka_config.topics = [MagicMock(name="topic1"), MagicMock(name="topic2")]
        # setup ELK configuration
        elk_container = MagicMock()
        elk_container.physical_host_ip = "192.168.1.10"
        elk_container.docker_gw_bridge_ip = "172.17.0.4"
        elk_config = MagicMock()
        emulation_env_config.elk_config = elk_config
        emulation_env_config.elk_config.container = elk_container
        emulation_env_config.elk_config.container.get_ips.return_value = ["192.168.1.3"]
        emulation_env_config.elk_config.kibana_port = 5601
        emulation_env_config.elk_config.elastic_port = 9200
        # setup SDN controller configuration
        sdn_container = MagicMock()
        sdn_container.physical_host_ip = "192.168.1.11"
        sdn_container.docker_gw_bridge_ip = "172.17.0.5"
        sdn_controller_config = MagicMock()
        emulation_env_config.sdn_controller_config = sdn_controller_config
        emulation_env_config.sdn_controller_config.container = sdn_container
        # setup host_manager configuration
        host_manager_config = MagicMock()
        emulation_env_config.host_manager_config = host_manager_config
        emulation_env_config.host_manager_config.host_manager_port = 12345
        # setup base_config
        beats_config = MagicMock()
        emulation_env_config.beats_config = beats_config
        emulation_env_config.beats_config.num_elastic_shards = 5
        emulation_env_config.beats_config.reload_enabled = True
        # setup node_beats_config
        node_beats_config = MagicMock()
        node_beats_config.log_files_paths = ["/var/log/syslog"]
        node_beats_config.filebeat_modules = ["system"]
        node_beats_config.kafka_input = True
        emulation_env_config.beats_config.get_node_beats_config_by_ips.return_value = (node_beats_config)
        emulation_env_config.execution_id = "execution_id"
        emulation_env_config.name = "emulation_name"
        self.emulation_env_config = emulation_env_config
        self.logger = MagicMock(spec=logging.Logger)

    @patch("csle_common.controllers.host_controller.HostController.start_host_manager")
    def test_start_host_managers(self, mock_start_host_manager) -> None:
        """
        Test start_host_managers

        :param mock_start_host_manager: mock start_host_manager
        :return: None
        """
        HostController.start_host_managers(self.emulation_env_config, self.logger)
        mock_start_host_manager.assert_called()
        assert mock_start_host_manager.call_count == 5

    @patch("csle_common.controllers.host_controller.HostController.get_host_monitor_thread_status_by_port_and_ip")
    @patch("csle_common.util.emulation_util.EmulationUtil.connect_admin")
    @patch("csle_common.util.emulation_util.EmulationUtil.execute_ssh_cmd")
    @patch("time.sleep", return_value=None)
    def test_start_host_manager(self, mock_sleep, mock_execute_ssh_cmd, mock_connect_admin,
                                mock_get_host_monitor_thread_status) -> None:
        """
        Test start_host_manager

        :param mock_sleep: mock sleep
        :param mock_execute_ssh_cmd: mock execute_ssh_cmd
        :param mock_connect_admin: mock connect_admin
        :param mock_get_host_monitor_thread_status: mock get_host_monitor_thread_status
        :return: None
        """
        mock_get_host_monitor_thread_status.return_value = None
        mock_connect_admin.return_value = None
        mock_execute_ssh_cmd.return_value = ("output", "error", 0)
        emulation_env_config = MagicMock(spec=EmulationEnvConfig)
        emulation_env_config.get_connection.return_value = MagicMock()
        host_manager_config = MagicMock()
        emulation_env_config.host_manager_config = host_manager_config
        emulation_env_config.host_manager_config.host_manager_port = 8080
        emulation_env_config.host_manager_config.host_manager_log_dir = ("/var/log/host_manager")
        emulation_env_config.host_manager_config.host_manager_log_file = ("host_manager.log")
        emulation_env_config.host_manager_config.host_manager_max_workers = 4
        containers_config = MagicMock()
        emulation_env_config.containers_config = containers_config
        containers_config.get_container_from_ip.return_value = None
        logger = MagicMock(spec=logging.Logger)
        ip = "172.17.0.1"
        # First call, should start it
        HostController.start_host_manager(emulation_env_config, ip, logger)
        host_monitor_dto = MagicMock()
        host_monitor_dto.monitor_running = False
        mock_get_host_monitor_thread_status.return_value = host_monitor_dto
        # Second call, should not start it
        HostController.start_host_manager(emulation_env_config, ip, logger)

        # Assert only started once
        mock_connect_admin.assert_called_once_with(emulation_env_config=emulation_env_config, ip=ip)
        mock_sleep.assert_called_once_with(2)

    @patch("csle_common.controllers.host_controller.HostController.stop_host_manager")
    def test_stop_host_managers(self, mock_stop_host_manager) -> None:
        """
        Test stop_host_managers

        :param mock_stop_host_manager: mock stop_host_manager
        :return: None
        """
        physical_host_ip = "192.168.1.10"
        HostController.stop_host_managers(self.emulation_env_config, physical_host_ip)
        mock_stop_host_manager.assert_called()
        assert mock_stop_host_manager.call_count == 3
        mock_stop_host_manager.reset_mock()
        physical_host_ip = "192.168.1.11"
        HostController.stop_host_managers(self.emulation_env_config, physical_host_ip)
        assert mock_stop_host_manager.call_count == 2

    @patch("csle_common.controllers.host_controller.EmulationUtil.connect_admin")
    @patch("csle_common.controllers.host_controller.Logger.__call__")
    @patch("csle_common.controllers.host_controller.EmulationUtil.execute_ssh_cmd")
    def test_stop_host_manager(self, mock_execute_ssh_cmd, mock_logger, mock_connect_admin) -> None:
        """
        Test stop_host_manager method

        :param mock_execute_ssh_cmd: mock execute_ssh_cmd
        :param mock_logger: mock logger
        :param mock_connect_admin: mock connect_admin
        :return: None
        """
        mock_logger_instance = MagicMock()
        mock_logger.return_value = mock_logger_instance
        mock_get_logger = mock_logger_instance.get_logger()
        mock_execute_ssh_cmd.return_value = ("output", "error", 0)
        ip = "172.17.0.2"
        HostController.stop_host_manager(emulation_env_config=self.emulation_env_config, ip=ip)
        mock_connect_admin.assert_called_once_with(emulation_env_config=self.emulation_env_config, ip=ip)
        mock_get_logger.info.assert_called_once_with(f"Stopping host manager on node {ip}")
        expected_cmd = (constants.COMMANDS.SUDO + constants.COMMANDS.SPACE_DELIM + constants.COMMANDS.PKILL +
                        constants.COMMANDS.SPACE_DELIM + constants.TRAFFIC_COMMANDS.HOST_MANAGER_FILE_NAME)
        mock_execute_ssh_cmd.assert_called_once_with(cmd=expected_cmd,
                                                     conn=self.emulation_env_config.get_connection(ip))

    @patch("csle_common.controllers.host_controller.HostController.start_host_monitor_thread")
    def test_start_host_monitor_threads(self, mock_start_host_monitor_thread) -> None:
        """
        Test start_host_monitor_threads method

        :param mock_start_host_monitor_threads: mock start_host_monitor_threads
        :return: None
        """
        physical_server_ip = "192.168.1.10"
        HostController.start_host_monitor_threads(
            emulation_env_config=self.emulation_env_config, physical_server_ip=physical_server_ip, logger=self.logger)
        assert mock_start_host_monitor_thread.call_count == 3
        mock_start_host_monitor_thread.reset_mock()
        physical_server_ip = "192.168.1.11"
        HostController.start_host_monitor_threads(emulation_env_config=self.emulation_env_config,
                                                  physical_server_ip=physical_server_ip, logger=self.logger)
        assert mock_start_host_monitor_thread.call_count == 2

    @patch("csle_common.controllers.host_controller.HostController.start_filebeat")
    def test_start_filebeats(self, mock_start_filebeat) -> None:
        """
        Test start_filebeats method

        :param mock_start_filebeat: mock start_filebeat
        :return: None
        """
        physical_server_ip = "192.168.1.10"
        HostController.start_filebeats(
            emulation_env_config=self.emulation_env_config,
            physical_server_ip=physical_server_ip,
            logger=self.logger,
            initial_start=True,
        )
        assert mock_start_filebeat.call_count == 3

    @patch("csle_common.controllers.host_controller.HostController.start_packetbeat")
    def test_start_packetbeats(self, mock_start_packetbeat) -> None:
        """
        Test start_packetbeats method

        :param mock_start_packetbeat: mock start_packetbeat
        :return: None
        """
        physical_server_ip = "192.168.1.10"
        HostController.start_packetbeats(
            emulation_env_config=self.emulation_env_config,
            physical_server_ip=physical_server_ip,
            logger=self.logger,
            initial_start=True,
        )
        assert mock_start_packetbeat.call_count == 3

    @patch("csle_common.controllers.host_controller.HostController.start_metricbeat")
    def test_start_metricbeats(self, mock_start_metricbeat) -> None:
        """
        Test start_metricbeats method

        :param mock_start_metricbeat: mock start_metricbeat
        :return: None
        """
        physical_server_ip = "192.168.1.10"
        HostController.start_metricbeats(
            emulation_env_config=self.emulation_env_config,
            physical_server_ip=physical_server_ip,
            logger=self.logger,
            initial_start=True,
        )
        assert mock_start_metricbeat.call_count == 3

    @patch("csle_common.controllers.host_controller.HostController.start_heartbeat")
    def test_start_heartbeats(self, mock_start_heartbeat) -> None:
        """
        Test start_heartbeats method

        :param mock_start_heartbeat: mock start_heartbeat
        :return: None
        """
        physical_server_ip = "192.168.1.10"
        HostController.start_heartbeats(
            emulation_env_config=self.emulation_env_config,
            physical_server_ip=physical_server_ip,
            logger=self.logger,
            initial_start=True,
        )
        assert mock_start_heartbeat.call_count == 3

    @patch("csle_common.controllers.host_controller.HostController.stop_filebeat")
    def test_stop_filebeats(self, mock_stop_filebeat) -> None:
        """
        Test stop_filebeats method

        :param mock_stop_filebeat: mock stop_filebeat
        :return: None
        """
        physical_server_ip = "192.168.1.10"
        HostController.stop_filebeats(
            emulation_env_config=self.emulation_env_config,
            physical_server_ip=physical_server_ip,
            logger=self.logger,
        )
        assert mock_stop_filebeat.call_count == 3

    @patch("csle_common.controllers.host_controller.HostController.stop_packetbeat")
    def test_stop_packetbeats(self, mock_stop_packetbeat) -> None:
        """
        Test stop_packetbeats method

        :param mock_stop_packetbeat: mock stop_packetbeat
        :return: None
        """
        physical_server_ip = "192.168.1.10"
        HostController.stop_packetbeats(
            emulation_env_config=self.emulation_env_config,
            physical_server_ip=physical_server_ip,
            logger=self.logger,
        )
        assert mock_stop_packetbeat.call_count == 3

    @patch("csle_common.controllers.host_controller.HostController.stop_metricbeat")
    def test_stop_metricbeats(self, mock_stop_metricbeat) -> None:
        """
        Test stop_metricbeats method

        :param mock_stop_metricbeat: mock stop_metricbeat
        :return: None
        """
        physical_server_ip = "192.168.1.10"
        HostController.stop_metricbeats(
            emulation_env_config=self.emulation_env_config,
            physical_server_ip=physical_server_ip,
            logger=self.logger,
        )
        assert mock_stop_metricbeat.call_count == 3

    @patch("csle_common.controllers.host_controller.HostController.stop_heartbeat")
    def test_stop_heartbeats(self, mock_stop_heartbeat) -> None:
        """
        Test stop_heartbeats method

        :param mock_stop_heartbeat: mock stop_heartbeat
        :return: None
        """
        physical_server_ip = "192.168.1.10"
        HostController.stop_heartbeats(
            emulation_env_config=self.emulation_env_config,
            physical_server_ip=physical_server_ip,
            logger=self.logger,
        )
        assert mock_stop_heartbeat.call_count == 3

    @patch("csle_common.controllers.host_controller.HostController.config_filebeat")
    def test_config_filebeats(self, mock_config_filebeat) -> None:
        """
        Test config_filebeats method

        :param mock_config_filebeat: mock config_filebeat
        :return: None
        """
        physical_server_ip = "192.168.1.10"
        HostController.config_filebeats(
            emulation_env_config=self.emulation_env_config,
            physical_server_ip=physical_server_ip,
            logger=self.logger,
        )
        assert mock_config_filebeat.call_count == 3

    @patch("csle_common.controllers.host_controller.HostController.config_packetbeat")
    def test_config_packetbeats(self, mock_config_packetbeat) -> None:
        """
        Test config_packetbeats method
        :param mock_config_packetbeat: mock config_packetbeat
        :return: None
        """
        physical_server_ip = "192.168.1.10"
        HostController.config_packetbeats(
            emulation_env_config=self.emulation_env_config,
            physical_server_ip=physical_server_ip,
            logger=self.logger,
        )
        assert mock_config_packetbeat.call_count == 3

    @patch("csle_common.controllers.host_controller.HostController.config_metricbeat")
    def test_config_metricbeats(self, mock_config_metricbeat) -> None:
        """
        Test config_metricbeats method

        :param mock_config_metricbeat: mock config_metricbeat
        :return: None
        """
        physical_server_ip = "192.168.1.10"
        HostController.config_metricbeats(
            emulation_env_config=self.emulation_env_config,
            physical_server_ip=physical_server_ip,
            logger=self.logger,
        )
        assert mock_config_metricbeat.call_count == 3

    @patch("csle_common.controllers.host_controller.HostController.config_heartbeat")
    def test_config_heartbeats(self, mock_config_heartbeat) -> None:
        """
        Test config_heartbeats method

        :param mock_config_heartbeat: mock config_heartbeat
        :return: None
        """
        physical_server_ip = "192.168.1.10"
        HostController.config_heartbeats(
            emulation_env_config=self.emulation_env_config,
            physical_server_ip=physical_server_ip,
            logger=self.logger,
        )
        assert mock_config_heartbeat.call_count == 3

    @patch("csle_common.controllers.host_controller.HostController.start_host_manager")
    @patch("csle_common.controllers.host_controller.HostController.get_host_monitor_thread_status_by_port_and_ip")
    @patch("csle_common.controllers.host_controller.grpc.insecure_channel")
    @patch("csle_common.controllers.host_controller.csle_collector.host_manager.query_host_manager.start_host_monitor")
    @patch("csle_common.controllers.host_controller.csle_collector.host_manager.host_manager_pb2_grpc.HostManagerStub")
    def test_start_host_monitor_thread(
            self, mock_HostManagerStub, mock_start_host_monitor, mock_insecure_channel,
            mock_get_host_monitor_thread_status, mock_start_host_manager) -> None:
        """
        Test start_host_monitor_thread method

        :param mock_HostManagerStub: mock HostManagerStub
        :param mock_start_host_monitor_thread: mock start_host_monitor_thread
        :param mock_insecure_channel: mock insecure_channel
        :param mock_get_host_monitor_thread_status: mock get_host_monitor_thread_status
        :param mock_start_host_manager: mock start_host_manager

        :return: None
        """
        # Mock the host_monitor_dto to return monitor_running as False
        host_monitor_dto = MagicMock()
        host_monitor_dto.monitor_running = False
        mock_get_host_monitor_thread_status.return_value = host_monitor_dto
        # Mock the gRPC channel and stub
        mock_channel = MagicMock()
        mock_insecure_channel.return_value.__enter__.return_value = mock_channel
        mock_stub = MagicMock()
        mock_HostManagerStub.return_value = mock_stub
        ip = "172.17.0.2"
        HostController.start_host_monitor_thread(emulation_env_config=self.emulation_env_config, ip=ip,
                                                 logger=self.logger)
        mock_start_host_manager.assert_called_once_with(emulation_env_config=self.emulation_env_config, ip=ip,
                                                        logger=self.logger)
        mock_get_host_monitor_thread_status.assert_called_once_with(ip=ip, port=12345)
        self.logger.info.assert_called_once_with(f"Host monitor thread is not running on {ip}, starting it.")
        mock_insecure_channel.assert_called_once_with(f"{ip}:12345", options=constants.GRPC_SERVERS.GRPC_OPTIONS)
        mock_HostManagerStub.assert_called_once_with(mock_channel)
        mock_start_host_monitor.assert_called_once_with(
            stub=mock_stub,
            kafka_ip="192.168.1.2",
            kafka_port=9092,
            time_step_len_seconds=10,
        )

    @patch("csle_common.controllers.host_controller.HostController.start_host_manager")
    @patch("csle_common.controllers.host_controller.HostController.get_host_monitor_thread_status_by_port_and_ip")
    @patch("csle_common.controllers.host_controller.grpc.insecure_channel")
    @patch("csle_common.controllers.host_controller.csle_collector.host_manager.query_host_manager.start_filebeat")
    @patch("csle_common.controllers.host_controller.csle_collector.host_manager.host_manager_pb2_grpc.HostManagerStub")
    def test_start_filebeat(
            self, mock_HostManagerStub, mock_start_filebeat, mock_insecure_channel, mock_get_host_monitor_thread_status,
            mock_start_host_manager) -> None:
        """
        Test start_filebeat method

        :param mock_HostManagerStub: mock HostManagerStub
        :param mock_start_filebeat: mock start_filebeat
        :param mock_insecure_channel: mock insecure_channel
        :param mock_get_host_monitor_thread_status: mock get_host_monitor_thread_status
        :param mock_start_host_manager: mock start_host_manager
        :return: None
        """
        beats_config = MagicMock()
        self.emulation_env_config.beats_config = beats_config
        self.emulation_env_config.beats_config.get_node_beats_config_by_ips.return_value = MagicMock(
            start_filebeat_automatically=True)
        host_monitor_dto = MagicMock()
        host_monitor_dto.filebeat_running = False
        mock_get_host_monitor_thread_status.return_value = host_monitor_dto
        mock_channel = MagicMock()
        mock_insecure_channel.return_value.__enter__.return_value = mock_channel
        mock_stub = MagicMock()
        mock_HostManagerStub.return_value = mock_stub
        ips = ["172.17.0.2", "192.168.1.2"]
        HostController.start_filebeat(emulation_env_config=self.emulation_env_config, ips=ips, logger=self.logger,
                                      initial_start=True)
        mock_start_host_manager.assert_called_once_with(emulation_env_config=self.emulation_env_config, ip=ips[0],
                                                        logger=self.logger)
        self.emulation_env_config.beats_config.get_node_beats_config_by_ips.assert_called_once_with(ips=ips)
        mock_get_host_monitor_thread_status.assert_called_once_with(ip=ips[0], port=12345)
        self.logger.info.assert_called_once_with(f"Filebeat is not running on {ips[0]}, starting it.")
        mock_insecure_channel.assert_called_once_with(f"{ips[0]}:12345", options=constants.GRPC_SERVERS.GRPC_OPTIONS)
        mock_HostManagerStub.assert_called_once_with(mock_channel)
        mock_start_filebeat.assert_called_once_with(stub=mock_stub)

    @patch("csle_common.controllers.host_controller.HostController.start_host_manager")
    @patch("csle_common.controllers.host_controller.HostController.get_host_monitor_thread_status_by_port_and_ip")
    @patch("csle_common.controllers.host_controller.grpc.insecure_channel")
    @patch("csle_common.controllers.host_controller.csle_collector.host_manager.query_host_manager.start_packetbeat")
    @patch("csle_common.controllers.host_controller.csle_collector.host_manager.host_manager_pb2_grpc.HostManagerStub")
    def test_start_packetbeat(
            self, mock_HostManagerStub, mock_start_packetbeat, mock_insecure_channel,
            mock_get_host_monitor_thread_status,
            mock_start_host_manager) -> None:
        """
        Test start_packetbeat method

        :param mock_HostManagerStub: mock HostManagerStub
        :param mock_start_packetbeat: mock start_packetbeat
        :param mock_insecure_channel: mock insecure_channel
        :param mock_get_host_monitor_thread_status: mock get_host_monitor_thread_status
        :param mock_start_host_manager: mock start_host_manager
        :return: None
        """
        beats_config = MagicMock()
        self.emulation_env_config.beats_config = beats_config
        self.emulation_env_config.beats_config.get_node_beats_config_by_ips.return_value = MagicMock(
            start_packetbeat_automatically=True)
        host_monitor_dto = MagicMock()
        host_monitor_dto.packetbeat_running = False
        mock_get_host_monitor_thread_status.return_value = host_monitor_dto
        mock_channel = MagicMock()
        mock_insecure_channel.return_value.__enter__.return_value = mock_channel
        mock_stub = MagicMock()
        mock_HostManagerStub.return_value = mock_stub
        ips = ["172.17.0.2", "192.168.1.2"]
        HostController.start_packetbeat(emulation_env_config=self.emulation_env_config, ips=ips, logger=self.logger,
                                        initial_start=True)
        mock_start_host_manager.assert_called_once_with(emulation_env_config=self.emulation_env_config, ip=ips[0],
                                                        logger=self.logger)
        self.emulation_env_config.beats_config.get_node_beats_config_by_ips.assert_called_once_with(ips=ips)
        mock_get_host_monitor_thread_status.assert_called_once_with(ip=ips[0], port=12345)
        self.logger.info.assert_called_once_with(f"Packetbeat is not running on {ips[0]}, starting it.")
        mock_insecure_channel.assert_called_once_with(f"{ips[0]}:12345", options=constants.GRPC_SERVERS.GRPC_OPTIONS)
        mock_HostManagerStub.assert_called_once_with(mock_channel)
        mock_start_packetbeat.assert_called_once_with(stub=mock_stub)

    @patch("csle_common.controllers.host_controller.HostController.start_host_manager")
    @patch("csle_common.controllers.host_controller.HostController.get_host_monitor_thread_status_by_port_and_ip")
    @patch("csle_common.controllers.host_controller.grpc.insecure_channel")
    @patch("csle_common.controllers.host_controller.csle_collector.host_manager.query_host_manager.start_metricbeat")
    @patch("csle_common.controllers.host_controller.csle_collector.host_manager.host_manager_pb2_grpc.HostManagerStub")
    def test_start_metricbeat(
            self, mock_HostManagerStub, mock_start_metricbeat, mock_insecure_channel,
            mock_get_host_monitor_thread_status,
            mock_start_host_manager) -> None:
        """
        Test start_metricbeat method

        :param mock_HostManagerStub: mock HostManagerStub
        :param mock_start_metricbeat: mock start_metricbeat
        :param mock_insecure_channel: mock insecure_channel
        :param mock_get_host_monitor_thread_status: mock get_host_monitor_thread_status
        :param mock_start_host_manager: mock start_host_manager
        :return: None
        """
        beats_config = MagicMock()
        self.emulation_env_config.beats_config = beats_config
        self.emulation_env_config.beats_config.get_node_beats_config_by_ips.return_value = MagicMock(
            start_metricbeat_automatically=True)
        host_monitor_dto = MagicMock()
        host_monitor_dto.metricbeat_running = False
        mock_get_host_monitor_thread_status.return_value = host_monitor_dto
        mock_channel = MagicMock()
        mock_insecure_channel.return_value.__enter__.return_value = mock_channel
        mock_stub = MagicMock()
        mock_HostManagerStub.return_value = mock_stub
        ips = ["172.17.0.2", "192.168.1.2"]
        HostController.start_metricbeat(emulation_env_config=self.emulation_env_config, ips=ips, logger=self.logger,
                                        initial_start=True)
        mock_start_host_manager.assert_called_once_with(emulation_env_config=self.emulation_env_config, ip=ips[0],
                                                        logger=self.logger)
        self.emulation_env_config.beats_config.get_node_beats_config_by_ips.assert_called_once_with(ips=ips)
        mock_get_host_monitor_thread_status.assert_called_once_with(ip=ips[0], port=12345)
        self.logger.info.assert_called_once_with(f"Metricbeat is not running on {ips[0]}, starting it.")
        mock_insecure_channel.assert_called_once_with(f"{ips[0]}:12345", options=constants.GRPC_SERVERS.GRPC_OPTIONS)
        mock_HostManagerStub.assert_called_once_with(mock_channel)
        mock_start_metricbeat.assert_called_once_with(stub=mock_stub)

    @patch("csle_common.controllers.host_controller.HostController.start_host_manager")
    @patch("csle_common.controllers.host_controller.HostController.get_host_monitor_thread_status_by_port_and_ip")
    @patch("csle_common.controllers.host_controller.grpc.insecure_channel")
    @patch("csle_common.controllers.host_controller.csle_collector.host_manager.query_host_manager.start_heartbeat")
    @patch("csle_common.controllers.host_controller.csle_collector.host_manager.host_manager_pb2_grpc.HostManagerStub")
    def test_start_heartbeat(self, mock_HostManagerStub, mock_start_heartbeat, mock_insecure_channel,
                             mock_get_host_monitor_thread_status, mock_start_host_manager) -> None:
        """
        Test start_heartbeat method

        :param mock_HostManagerStub: mock HostManagerStub
        :param mock_start_heartbeat: mock start_heartbeat
        :param mock_insecure_channel: mock insecure_channel
        :param mock_get_host_monitor_thread_status: mock get_host_monitor_thread_status
        :param mock_start_host_manager: mock start_host_manager
        :return: None
        """
        beats_config = MagicMock()
        self.emulation_env_config.beats_config = beats_config
        self.emulation_env_config.beats_config.get_node_beats_config_by_ips.return_value = MagicMock(
            start_heartbeat_automatically=True)
        host_monitor_dto = MagicMock()
        host_monitor_dto.heartbeat_running = False
        mock_get_host_monitor_thread_status.return_value = host_monitor_dto
        mock_channel = MagicMock()
        mock_insecure_channel.return_value.__enter__.return_value = mock_channel
        mock_stub = MagicMock()
        mock_HostManagerStub.return_value = mock_stub
        ips = ["172.17.0.2", "192.168.1.2"]
        HostController.start_heartbeat(emulation_env_config=self.emulation_env_config, ips=ips, logger=self.logger,
                                       initial_start=True)
        mock_start_host_manager.assert_called_once_with(emulation_env_config=self.emulation_env_config, ip=ips[0],
                                                        logger=self.logger)
        self.emulation_env_config.beats_config.get_node_beats_config_by_ips.assert_called_once_with(ips=ips)
        mock_get_host_monitor_thread_status.assert_called_once_with(ip=ips[0], port=12345)
        self.logger.info.assert_called_once_with(f"Heartbeat is not running on {ips[0]}, starting it.")
        mock_insecure_channel.assert_called_once_with(f"{ips[0]}:12345", options=constants.GRPC_SERVERS.GRPC_OPTIONS)
        mock_HostManagerStub.assert_called_once_with(mock_channel)
        mock_start_heartbeat.assert_called_once_with(stub=mock_stub)

    @patch("csle_common.controllers.host_controller.HostController.start_spark")
    def test_start_sparks(self, mock_start_spark) -> None:
        """
        Test start_sparks method

        :param mock_start_spark: mock start_spark
        :return: None
        """
        constants.CONTAINER_IMAGES.SPARK_IMAGES = ["spark-container-1"]
        container1 = self.emulation_env_config.containers_config.containers[0]
        physical_server_ip = "192.168.1.10"
        HostController.start_sparks(
            emulation_env_config=self.emulation_env_config, physical_server_ip=physical_server_ip, logger=self.logger)
        self.logger.info.assert_called_once_with(f"Starting Spark on IP: {container1.docker_gw_bridge_ip} "
                                                 f"({container1.get_ips()[0]}, {container1.get_full_name()})")
        mock_start_spark.assert_called_once_with(
            emulation_env_config=self.emulation_env_config, ips=[container1.docker_gw_bridge_ip], logger=self.logger)
        # Reset mocks to clear previous call history
        self.logger.reset_mock()
        mock_start_spark.reset_mock()
        physical_server_ip = "192.168.1.11"
        HostController.start_sparks(
            emulation_env_config=self.emulation_env_config, physical_server_ip=physical_server_ip, logger=self.logger)
        self.logger.info.assert_not_called()
        mock_start_spark.assert_not_called()

    @patch("csle_common.controllers.host_controller.HostController.stop_spark")
    def test_stop_sparks(self, mock_stop_spark) -> None:
        """
        Test stop_sparks method

        :param mock_stop_spark: mock stop_sparks
        :return: None
        """
        constants.CONTAINER_IMAGES.SPARK_IMAGES = ["spark-container-1"]
        container1 = self.emulation_env_config.containers_config.containers[0]
        physical_server_ip = "192.168.1.10"
        HostController.stop_sparks(emulation_env_config=self.emulation_env_config,
                                   physical_server_ip=physical_server_ip,
                                   logger=self.logger)
        self.logger.info.assert_called_once_with(f"Stopping Spark on IP: {container1.docker_gw_bridge_ip} "
                                                 f"({container1.get_ips()[0]}, {container1.get_full_name()})")
        mock_stop_spark.assert_called_once_with(
            emulation_env_config=self.emulation_env_config, ips=[container1.docker_gw_bridge_ip], logger=self.logger)
        # Reset mocks to clear previous call history
        self.logger.reset_mock()
        mock_stop_spark.reset_mock()
        physical_server_ip = "192.168.1.11"
        HostController.stop_sparks(
            emulation_env_config=self.emulation_env_config, physical_server_ip=physical_server_ip, logger=self.logger)
        self.logger.info.assert_not_called()
        mock_stop_spark.assert_not_called()

    @patch("csle_common.controllers.host_controller.HostController.start_host_manager")
    @patch("csle_common.controllers.host_controller.grpc.insecure_channel")
    @patch("csle_common.controllers.host_controller.csle_collector.host_manager.query_host_manager.start_spark")
    @patch("csle_common.controllers.host_controller.csle_collector.host_manager.host_manager_pb2_grpc.HostManagerStub")
    def test_start_spark(self, mock_HostManagerStub, mock_start_spark, mock_insecure_channel, mock_start_host_manager) \
            -> None:
        """
        Test start_spark method

        :param mock_HostManagerStub: mock HostManagerStub
        :param mock_start_spark: mock start_spark
        :param mock_insecure_channel: mock insecure_channel
        :param mock_start_host_manager: mock start_host_manager
        :return: None
        """
        mock_channel = MagicMock()
        mock_insecure_channel.return_value.__enter__.return_value = mock_channel
        mock_stub = MagicMock()
        mock_HostManagerStub.return_value = mock_stub
        ips = ["172.17.0.2", "192.168.1.2"]
        HostController.start_spark(emulation_env_config=self.emulation_env_config, ips=ips, logger=self.logger)
        mock_start_host_manager.assert_called_once_with(
            emulation_env_config=self.emulation_env_config, ip=ips[0], logger=self.logger)
        mock_insecure_channel.assert_called_once_with(f"{ips[0]}:12345", options=constants.GRPC_SERVERS.GRPC_OPTIONS)
        mock_HostManagerStub.assert_called_once_with(mock_channel)
        mock_start_spark.assert_called_once_with(stub=mock_stub)

    @patch("csle_common.controllers.host_controller.HostController.start_host_manager")
    @patch("csle_common.controllers.host_controller.grpc.insecure_channel")
    @patch("csle_common.controllers.host_controller.csle_collector.host_manager.query_host_manager.stop_spark")
    @patch("csle_common.controllers.host_controller.csle_collector.host_manager.host_manager_pb2_grpc.HostManagerStub")
    def test_stop_spark(self, mock_HostManagerStub, mock_stop_spark, mock_insecure_channel, mock_start_host_manager) \
            -> None:
        """
        Test stop_spark method

        :param mock_HostManagerStub: mock HostManagerStub
        :param mock_stop_spark: mock stop_spark
        :param mock_insecure_channel: mock insecure_channel
        :param mock_start_host_manager: mock start_host_manager
        :return: None
        """
        mock_channel = MagicMock()
        mock_insecure_channel.return_value.__enter__.return_value = mock_channel
        mock_stub = MagicMock()
        mock_HostManagerStub.return_value = mock_stub
        ips = ["172.17.0.2", "192.168.1.2"]
        HostController.stop_spark(emulation_env_config=self.emulation_env_config, ips=ips, logger=self.logger)
        mock_start_host_manager.assert_called_once_with(
            emulation_env_config=self.emulation_env_config, ip=ips[0], logger=self.logger)
        mock_insecure_channel.assert_called_once_with(f"{ips[0]}:12345", options=constants.GRPC_SERVERS.GRPC_OPTIONS)
        mock_HostManagerStub.assert_called_once_with(mock_channel)
        mock_stop_spark.assert_called_once_with(stub=mock_stub)

    @patch("csle_common.controllers.host_controller.HostController.start_host_manager")
    @patch("csle_common.controllers.host_controller.grpc.insecure_channel")
    @patch("csle_common.controllers.host_controller.csle_collector.host_manager.query_host_manager.config_filebeat")
    @patch("csle_common.controllers.host_controller.csle_collector.host_manager.host_manager_pb2_grpc.HostManagerStub")
    def test_config_filebeat(
            self, mock_HostManagerStub, mock_config_filebeat, mock_insecure_channel, mock_start_host_manager) -> None:
        """
        Test config_filebeat method

        :param mock_HostManagerStub: mock HostManagerStub
        :param mock_config_filebeat: mock stop_config_filebeat
        :param mock_insecure_channel: mock insecure_channel
        :param mock_start_host_manager: mock start_host_manager
        :return: None
        """
        # Mock the gRPC channel and stub
        mock_channel = MagicMock()
        mock_insecure_channel.return_value.__enter__.return_value = mock_channel
        mock_stub = MagicMock()
        mock_HostManagerStub.return_value = mock_stub
        # Call the function
        container = self.emulation_env_config.containers_config.containers[0]
        HostController.config_filebeat(self.emulation_env_config, container, self.logger)
        mock_start_host_manager.assert_called_once_with(
            emulation_env_config=self.emulation_env_config, ip=container.docker_gw_bridge_ip, logger=self.logger)
        self.emulation_env_config.beats_config.get_node_beats_config_by_ips.assert_called_once_with(
            ips=container.get_ips())
        self.logger.info.assert_called_once_with(
            f"Configuring filebeat on {container.docker_gw_bridge_ip}, "
            f"{container.get_full_name()}, ips: {container.get_ips()}"
        )
        mock_insecure_channel.assert_called_once_with(
            f"{container.docker_gw_bridge_ip}:12345", options=constants.GRPC_SERVERS.GRPC_OPTIONS)
        mock_HostManagerStub.assert_called_once_with(mock_channel)
        mock_config_filebeat.assert_called()

    @patch("csle_common.controllers.host_controller.HostController.start_host_manager")
    @patch("csle_common.controllers.host_controller.grpc.insecure_channel")
    @patch("csle_common.controllers.host_controller.csle_collector.host_manager.query_host_manager.config_packetbeat")
    @patch("csle_common.controllers.host_controller.csle_collector.host_manager.host_manager_pb2_grpc.HostManagerStub")
    def test_config_packetbeat(
            self, mock_HostManagerStub, mock_config_packetbeat, mock_insecure_channel, mock_start_host_manager) -> None:
        """
        Test config_packetbeat method

        :param mock_HostManagerStub: mock HostManagerStub
        :param mock_config_packetbeat: mock stop_config_packetbeat
        :param mock_insecure_channel: mock insecure_channel
        :param mock_start_host_manager: mock start_host_manager
        :return: None
        """
        # Mock the gRPC channel and stub
        mock_channel = MagicMock()
        mock_insecure_channel.return_value.__enter__.return_value = mock_channel
        mock_stub = MagicMock()
        mock_HostManagerStub.return_value = mock_stub
        # Call the function
        container = self.emulation_env_config.containers_config.containers[0]
        HostController.config_packetbeat(self.emulation_env_config, container, self.logger)
        mock_start_host_manager.assert_called_once_with(
            emulation_env_config=self.emulation_env_config, ip=container.docker_gw_bridge_ip, logger=self.logger)
        self.logger.info.assert_called_once_with(
            f"Configuring packetbeat on {container.docker_gw_bridge_ip}, "
            f"{container.get_full_name()}, ips: {container.get_ips()}")
        mock_insecure_channel.assert_called_once_with(f"{container.docker_gw_bridge_ip}:12345",
                                                      options=constants.GRPC_SERVERS.GRPC_OPTIONS)
        mock_HostManagerStub.assert_called_once_with(mock_channel)
        mock_config_packetbeat.assert_called()

    @patch("csle_common.controllers.host_controller.HostController.start_host_manager")
    @patch("csle_common.controllers.host_controller.grpc.insecure_channel")
    @patch("csle_common.controllers.host_controller.csle_collector.host_manager.query_host_manager.config_metricbeat")
    @patch("csle_common.controllers.host_controller.csle_collector.host_manager.host_manager_pb2_grpc.HostManagerStub")
    def test_config_metricbeat(
            self, mock_HostManagerStub, mock_config_metricbeat, mock_insecure_channel, mock_start_host_manager) -> None:
        """
        Test config_metricbeat method

        :param mock_HostManagerStub: mock HostManagerStub
        :param mock_config_metricbeat: mock stop_config_metricbeat
        :param mock_insecure_channel: mock insecure_channel
        :param mock_start_host_manager: mock start_host_manager
        :return: None
        """
        # Mock the gRPC channel and stub
        mock_channel = MagicMock()
        mock_insecure_channel.return_value.__enter__.return_value = mock_channel
        mock_stub = MagicMock()
        mock_HostManagerStub.return_value = mock_stub
        # Call the function
        container = self.emulation_env_config.containers_config.containers[0]
        HostController.config_metricbeat(self.emulation_env_config, container, self.logger)
        mock_start_host_manager.assert_called_once_with(
            emulation_env_config=self.emulation_env_config, ip=container.docker_gw_bridge_ip, logger=self.logger)
        self.emulation_env_config.beats_config.get_node_beats_config_by_ips.assert_called_once_with(
            ips=container.get_ips())
        self.logger.info.assert_called_once_with(
            f"Configuring metricbeat on {container.docker_gw_bridge_ip}, "
            f"{container.get_full_name()}, ips: {container.get_ips()}")
        mock_insecure_channel.assert_called_once_with(
            f"{container.docker_gw_bridge_ip}:12345", options=constants.GRPC_SERVERS.GRPC_OPTIONS)
        mock_HostManagerStub.assert_called_once_with(mock_channel)
        mock_config_metricbeat.assert_called()

    @patch("csle_common.controllers.host_controller.HostController.start_host_manager")
    @patch("csle_common.controllers.host_controller.grpc.insecure_channel")
    @patch("csle_common.controllers.host_controller.csle_collector.host_manager.query_host_manager.config_heartbeat")
    @patch("csle_common.controllers.host_controller.csle_collector.host_manager.host_manager_pb2_grpc.HostManagerStub")
    def test_config_heartbeat(self, mock_HostManagerStub, mock_config_heartbeat, mock_insecure_channel,
                              mock_start_host_manager) -> None:
        """
        Test config_heartbeat method

        :param mock_HostManagerStub: mock HostManagerStub
        :param mock_config_heartbeat: mock stop_config_heartbeat
        :param mock_insecure_channel: mock insecure_channel
        :param mock_start_host_manager: mock start_host_manager
        :return: None
        """
        # Mock the gRPC channel and stub
        mock_channel = MagicMock()
        mock_insecure_channel.return_value.__enter__.return_value = mock_channel
        mock_stub = MagicMock()
        mock_HostManagerStub.return_value = mock_stub
        # Call the function
        container = self.emulation_env_config.containers_config.containers[0]
        HostController.config_heartbeat(self.emulation_env_config, container, self.logger)
        mock_start_host_manager.assert_called_once_with(
            emulation_env_config=self.emulation_env_config,
            ip=container.docker_gw_bridge_ip,
            logger=self.logger,
        )
        self.emulation_env_config.beats_config.get_node_beats_config_by_ips.assert_called_once_with(
            ips=container.get_ips())
        self.logger.info.assert_called_once_with(
            f"Configuring heartbeat on {container.docker_gw_bridge_ip}, "
            f"{container.get_full_name()}, ips: {container.get_ips()}")
        mock_insecure_channel.assert_called_once_with(
            f"{container.docker_gw_bridge_ip}:12345", options=constants.GRPC_SERVERS.GRPC_OPTIONS)
        mock_HostManagerStub.assert_called_once_with(mock_channel)
        mock_config_heartbeat.assert_called()

    @patch("csle_common.controllers.host_controller.HostController.stop_host_monitor_thread")
    def test_stop_host_monitor_threads(self, mock_stop_host_monitor_thread) -> None:
        """
        Test stop_host_monitor_threads method

        :param mock_stop_host_monitor_thread: mock stop_host_monitor_thread
        :return: None
        """
        physical_host_ip = "192.168.1.10"
        HostController.stop_host_monitor_threads(
            emulation_env_config=self.emulation_env_config, physical_host_ip=physical_host_ip, logger=self.logger)
        assert mock_stop_host_monitor_thread.call_count == 3

    @patch("csle_common.controllers.host_controller.HostController.start_host_manager")
    @patch("csle_common.controllers.host_controller.grpc.insecure_channel")
    @patch("csle_common.controllers.host_controller.csle_collector.host_manager.query_host_manager.stop_host_monitor")
    @patch("csle_common.controllers.host_controller.csle_collector.host_manager.host_manager_pb2_grpc.HostManagerStub")
    def test_stop_host_monitor_thread(self, mock_HostManagerStub, mock_stop_host_monitor_thread,
                                      mock_insecure_channel, mock_start_host_manager) -> None:
        """
        Test stop_host_monitor_thread method

        :param mock_HostManagerStub: mock HostManagerStub
        :param mock_stop_host_monitor_thread: mock stop_host_monitor_thread
        :param mock_insecure_channel: mock insecure_channel
        :param mock_start_host_manager: mock start_host_manager
        :return: None
        """
        mock_channel = MagicMock()
        mock_insecure_channel.return_value.__enter__.return_value = mock_channel
        mock_stub = MagicMock()
        mock_HostManagerStub.return_value = mock_stub
        ip = "172.17.0.2"
        HostController.stop_host_monitor_thread(emulation_env_config=self.emulation_env_config, ip=ip,
                                                logger=self.logger)
        mock_start_host_manager.assert_called_once_with(emulation_env_config=self.emulation_env_config, ip=ip,
                                                        logger=self.logger)
        mock_insecure_channel.assert_called_once_with(f"{ip}:12345", options=constants.GRPC_SERVERS.GRPC_OPTIONS)
        mock_HostManagerStub.assert_called_once_with(mock_channel)
        mock_stop_host_monitor_thread.assert_called_once_with(stub=mock_stub)

    @patch("csle_common.controllers.host_controller.HostController.start_host_manager")
    @patch("csle_common.controllers.host_controller.grpc.insecure_channel")
    @patch("csle_common.controllers.host_controller.csle_collector.host_manager.query_host_manager.stop_filebeat")
    @patch("csle_common.controllers.host_controller.csle_collector.host_manager.host_manager_pb2_grpc.HostManagerStub")
    def test_stop_filebeat(self, mock_HostManagerStub, mock_stop_filebeat, mock_insecure_channel,
                           mock_start_host_manager) -> None:
        """
        Test stop_filebeat method

        :param mock_HostManagerStub: mock HostManagerStub
        :param mock_stop_filebeat: mock stop_filebeat
        :param mock_insecure_channel: mock insecure_channel
        :param mock_start_host_manager: mock start_host_manager
        :return: None
        """
        mock_channel = MagicMock()
        mock_insecure_channel.return_value.__enter__.return_value = mock_channel
        mock_stub = MagicMock()
        mock_HostManagerStub.return_value = mock_stub
        ip = "172.17.0.2"
        HostController.stop_filebeat(emulation_env_config=self.emulation_env_config, ip=ip, logger=self.logger)
        mock_start_host_manager.assert_called_once_with(emulation_env_config=self.emulation_env_config, ip=ip,
                                                        logger=self.logger)
        mock_insecure_channel.assert_called_once_with(f"{ip}:12345", options=constants.GRPC_SERVERS.GRPC_OPTIONS)
        mock_HostManagerStub.assert_called_once_with(mock_channel)
        mock_stop_filebeat.assert_called_once_with(stub=mock_stub)

    @patch("csle_common.controllers.host_controller.HostController.start_host_manager")
    @patch("csle_common.controllers.host_controller.grpc.insecure_channel")
    @patch("csle_common.controllers.host_controller.csle_collector.host_manager.query_host_manager.stop_packetbeat")
    @patch("csle_common.controllers.host_controller.csle_collector.host_manager.host_manager_pb2_grpc.HostManagerStub")
    def test_stop_packetbeat(self, mock_HostManagerStub, mock_stop_packetbeat, mock_insecure_channel,
                             mock_start_host_manager) -> None:
        """
        Test stop_packetbeat method

        :param mock_HostManagerStub: mock HostManagerStub
        :param mock_stop_packetbeat: mock stop_packetbeat
        :param mock_insecure_channel: mock insecure_channel
        :param mock_start_host_manager: mock start_host_manager
        :return: None
        """
        mock_channel = MagicMock()
        mock_insecure_channel.return_value.__enter__.return_value = mock_channel
        mock_stub = MagicMock()
        mock_HostManagerStub.return_value = mock_stub
        ip = "172.17.0.2"
        HostController.stop_packetbeat(emulation_env_config=self.emulation_env_config, ip=ip, logger=self.logger)
        mock_start_host_manager.assert_called_once_with(emulation_env_config=self.emulation_env_config, ip=ip,
                                                        logger=self.logger)
        mock_insecure_channel.assert_called_once_with(f"{ip}:12345", options=constants.GRPC_SERVERS.GRPC_OPTIONS)
        mock_HostManagerStub.assert_called_once_with(mock_channel)
        mock_stop_packetbeat.assert_called_once_with(stub=mock_stub)

    @patch("csle_common.controllers.host_controller.HostController.start_host_manager")
    @patch("csle_common.controllers.host_controller.grpc.insecure_channel")
    @patch("csle_common.controllers.host_controller.csle_collector.host_manager.query_host_manager.stop_metricbeat")
    @patch("csle_common.controllers.host_controller.csle_collector.host_manager.host_manager_pb2_grpc.HostManagerStub")
    def test_stop_metricbeat(self, mock_HostManagerStub, mock_stop_metricbeat, mock_insecure_channel,
                             mock_start_host_manager) -> None:
        """
        Test stop_metricbeat method

        :param mock_HostManagerStub: mock HostManagerStub
        :param mock_stop_metricbeat: mock stop_metricbeat
        :param mock_insecure_channel: mock insecure_channel
        :param mock_start_host_manager: mock start_host_manager
        :return: None
        """
        mock_channel = MagicMock()
        mock_insecure_channel.return_value.__enter__.return_value = mock_channel
        mock_stub = MagicMock()
        mock_HostManagerStub.return_value = mock_stub
        ip = "172.17.0.2"
        HostController.stop_metricbeat(emulation_env_config=self.emulation_env_config, ip=ip, logger=self.logger)
        mock_start_host_manager.assert_called_once_with(emulation_env_config=self.emulation_env_config, ip=ip,
                                                        logger=self.logger)
        mock_insecure_channel.assert_called_once_with(f"{ip}:12345", options=constants.GRPC_SERVERS.GRPC_OPTIONS)
        mock_HostManagerStub.assert_called_once_with(mock_channel)
        mock_stop_metricbeat.assert_called_once_with(stub=mock_stub)

    @patch("csle_common.controllers.host_controller.HostController.start_host_manager")
    @patch("csle_common.controllers.host_controller.grpc.insecure_channel")
    @patch("csle_common.controllers.host_controller.csle_collector.host_manager.query_host_manager.stop_heartbeat")
    @patch("csle_common.controllers.host_controller.csle_collector.host_manager.host_manager_pb2_grpc.HostManagerStub")
    def test_stop_heartbeat(self, mock_HostManagerStub, mock_stop_heartbeat, mock_insecure_channel,
                            mock_start_host_manager) -> None:
        """
        Test stop_heartbeat method

        :param mock_HostManagerStub: mock HostManagerStub
        :param mock_stop_heartbeat: mock stop_heartbeat
        :param mock_insecure_channel: mock insecure_channel
        :param mock_start_host_manager: mock start_host_manager
        :return: None
        """
        mock_channel = MagicMock()
        mock_insecure_channel.return_value.__enter__.return_value = mock_channel
        mock_stub = MagicMock()
        mock_HostManagerStub.return_value = mock_stub
        ip = "172.17.0.2"
        HostController.stop_heartbeat(emulation_env_config=self.emulation_env_config, ip=ip, logger=self.logger)
        mock_start_host_manager.assert_called_once_with(emulation_env_config=self.emulation_env_config, ip=ip,
                                                        logger=self.logger)
        mock_insecure_channel.assert_called_once_with(f"{ip}:12345", options=constants.GRPC_SERVERS.GRPC_OPTIONS)
        mock_HostManagerStub.assert_called_once_with(mock_channel)
        mock_stop_heartbeat.assert_called_once_with(stub=mock_stub)

    @patch("csle_common.controllers.host_controller.HostController.start_host_managers")
    @patch("csle_common.controllers.host_controller.HostController.get_host_monitor_thread_status_by_port_and_ip")
    def test_get_host_monitor_threads_statuses(self, mock_get_status, mock_start_host_managers) -> None:
        """
        Test get_host_monitor_threads_statuses method

        :param mock_get_status: mock get_status method
        :param mock_start_host_managers: mock start_host_managers
        :return: None
        """
        status_mock = MagicMock()
        mock_get_status.return_value = status_mock
        physical_server_ip = "192.168.1.10"
        result = HostController.get_host_monitor_threads_statuses(
            emulation_env_config=self.emulation_env_config, physical_server_ip=physical_server_ip, logger=self.logger)
        mock_start_host_managers.assert_called_once_with(emulation_env_config=self.emulation_env_config,
                                                         logger=self.logger)
        assert mock_get_status.call_count == 3
        expected_result = [status_mock, status_mock, status_mock]
        assert result == expected_result

    @patch("csle_collector.host_manager.query_host_manager.get_host_status")
    @patch("csle_collector.host_manager.host_manager_pb2_grpc.HostManagerStub")
    @patch("grpc.insecure_channel")
    def test_get_host_monitor_thread_status_by_port_and_ip(self, mock_insecure_channel, mock_HostManagerStub,
                                                           mock_get_host_status) -> None:
        """
        Test get_host_monitor_thread_status_by_port_and_ip method

        :param mock_insecure_channel: mock insecure_channel
        :param mock_HostManagerStub: mock HostManagerStub
        :param mock_get_host_status: mock get_host_status
        :return: None
        """
        # Create a mock status to return
        status_mock = MagicMock()
        mock_get_host_status.return_value = status_mock
        mock_channel = MagicMock()
        mock_insecure_channel.return_value.__enter__.return_value = mock_channel
        mock_stub = MagicMock()
        mock_HostManagerStub.return_value = mock_stub
        ip = "172.17.0.2"
        port = 12345
        timeout = 5
        result = HostController.get_host_monitor_thread_status_by_port_and_ip(ip=ip, port=port, timeout=timeout)
        mock_insecure_channel.assert_called_once_with(f"{ip}:{port}", options=constants.GRPC_SERVERS.GRPC_OPTIONS)
        mock_HostManagerStub.assert_called_once_with(mock_channel)
        mock_get_host_status.assert_called_once_with(stub=mock_stub, timeout=timeout)
        assert result == status_mock

    def test_get_host_managers_ips(self) -> None:
        """
        Test the method for extracting the ips of the Host managers in a given emulation

        :return: None
        """
        result = HostController.get_host_managers_ips(self.emulation_env_config)
        expected_result = ["172.17.0.1", "172.17.0.2", "172.17.0.3", "172.17.0.4", "172.17.0.5"]
        assert result == expected_result

    def test_get_host_managers_ports(self) -> None:
        """
        Test the method for extracting the ports of the Host managers in a given emulation

        :return: None
        """
        result = HostController.get_host_managers_ports(self.emulation_env_config)
        expected_result = [12345, 12345, 12345, 12345, 12345]
        assert result == expected_result

    def test_get_host_managers_info(self) -> None:
        """
        Test the method for extracting the information of the Host managers for a given emulation

        :retur: None
        """
        active_ips = ["172.17.0.2", "172.17.0.3"]
        physical_host_ip = "192.168.1.1"
        result = HostController.get_host_managers_info(self.emulation_env_config, active_ips, self.logger,
                                                       physical_host_ip)
        assert result



import threading
import time
from confluent_kafka import Consumer, KafkaError, KafkaException
import csle_collector.constants.constants as collector_constants
from csle_collector.snort_ids_manager.dao.snort_ids_ip_alert_counters import SnortIdsIPAlertCounters
from csle_common.logging.log import Logger


class SnortIdsLogConsumerThread(threading.Thread):
    """
    Thread that polls the Snort IDS log to get the latest metrics
    """

    def __init__(self, host_ip: str,
                 kafka_server_ip: str, kafka_port: int, snort_ids_alert_counters: SnortIdsIPAlertCounters,
                 auto_offset_reset: str = "latest") -> None:
        """
        Initializes the thread

        :param host_ip: the ip of the host
        :param kafka_server_ip: the ip of the kafka server
        :param kafka_port: the port of the kafka server
        :param snort_ids_alert_counters: the alert counters to update
        :param auto_offset_reset: the offset for kafka to start reading from
        """
        threading.Thread.__init__(self)
        self.running = True
        self.host_ip = host_ip
        self.kafka_server_ip = kafka_server_ip
        self.kafka_port = kafka_port
        self.snort_ids_alert_counters = snort_ids_alert_counters
        self.ts = time.time()
        self.kafka_conf = {
            collector_constants.KAFKA.BOOTSTRAP_SERVERS_PROPERTY: f"{self.kafka_server_ip}:{self.kafka_port}",
            collector_constants.KAFKA.GROUP_ID_PROPERTY: f"ids_log_consumer_thread_{self.ts}",
            collector_constants.KAFKA.AUTO_OFFSET_RESET_PROPERTY: auto_offset_reset}
        self.consumer = Consumer(**self.kafka_conf)
        self.consumer.subscribe([collector_constants.KAFKA_CONFIG.SNORT_IDS_IP_LOG_TOPIC_NAME])

    def run(self) -> None:
        """
        Runs the thread

        :return: None
        """
        while self.running:
            msg = self.consumer.poll(timeout=5.0)
            if msg is not None:
                if msg.error():
                    if msg.error().code() == KafkaError._PARTITION_EOF:
                        Logger.__call__().get_logger.warning(
                            f"reached end of partition: {msg.topic(), msg.partition(), msg.offset()}")
                    elif msg.error():
                        raise KafkaException(msg.error())
                else:
                    self.snort_ids_alert_counters.update_with_kafka_record(record=msg.value().decode(), ip=self.host_ip)
        self.consumer.close()



import os
from csle_common.metastore.metastore_facade import MetastoreFacade
import csle_common.constants.constants as constants
from csle_common.dao.emulation_config.config import Config
from csle_common.logging.log import Logger
from csle_common.util.general_util import GeneralUtil


class ClusterUtil:
    """
    Class with utility related to cluster management
    """

    @staticmethod
    def am_i_leader(ip: str, config: Config) -> bool:
        """
        Checks if a given IP is leader or not

        :param ip: the ip to check
        :param config: the cluster configuration
        :return: True if leader, False otherwise
        """
        leader = False
        for node in config.cluster_config.cluster_nodes:
            if node.ip == ip:
                leader = node.leader
        return leader

    @staticmethod
    def get_leader_ip(config: Config) -> str:
        """
        Returns the IP of the leader node

        :param config: the cluster configuration
        :return: The IP address of the leader node
        """
        for node in config.cluster_config.cluster_nodes:
            if node.leader:
                return node.ip
        raise ValueError("No leader node found")

    @staticmethod
    def get_config() -> Config:
        """
        Gets the current cluster config from the metastore or from disk depending on if it is the leader node or
        not
        :return: the cluster config
        """
        config = MetastoreFacade.get_config(id=1)
        if config is None:
            config = Config.read_config_file()
        if config.localhost:
            ip = constants.COMMON.LOCALHOST_127_0_0_1
        else:
            ip = GeneralUtil.get_host_ip()
        constants.METADATA_STORE.HOST = ip
        constants.CLUSTER_CONFIG.IP = ip
        leader = ClusterUtil.am_i_leader(ip=ip, config=config)
        constants.CLUSTER_CONFIG.LEADER = leader
        if leader:
            config = Config.read_config_file()
            current_config = MetastoreFacade.get_config(id=1)
            if current_config is None:
                Logger.__call__().get_logger().info("Saving config to metastore")
                MetastoreFacade.save_config(config)
            else:
                MetastoreFacade.update_config(config=config, id=1)
        return config

    @staticmethod
    def set_config_parameters_from_config_file() -> None:
        """
        Reads the config file from $CSLE_HOME/config.json and initializes certain config parameters

        :return: None
        """
        if constants.CONFIG_FILE.CSLE_HOME_ENV_PARAM in os.environ:
            csle_home = os.environ[constants.CONFIG_FILE.CSLE_HOME_ENV_PARAM]
        else:
            raise Exception(f"The environment parameter {constants.CONFIG_FILE.CSLE_HOME_ENV_PARAM} is not set")
        config_file_path = f"{csle_home}{constants.COMMANDS.SLASH_DELIM}" \
                           f"{constants.CONFIG_FILE.CONFIG_FILE_NAME}"
        try:
            config = ClusterUtil.get_config()
            constants.CONFIG_FILE.PARSED_CONFIG = config
            constants.CSLE_ADMIN.MANAGEMENT_USER = config.management_admin_username_default
            constants.CSLE_ADMIN.MANAGEMENT_PW = config.management_admin_password_default
            constants.CSLE_ADMIN.MANAGEMENT_EMAIL = config.management_admin_email_default
            constants.CSLE_ADMIN.MANAGEMENT_ORGANIZATION = config.management_admin_organization_default
            constants.CSLE_ADMIN.MANAGEMENT_FIRST_NAME = config.management_admin_first_name_default
            constants.CSLE_ADMIN.MANAGEMENT_LAST_NAME = config.management_admin_last_name_default
            constants.CSLE_ADMIN.SSH_USER = config.ssh_admin_username
            constants.CSLE_ADMIN.SSH_PW = config.ssh_admin_password
            constants.CSLE_GUEST.MANAGEMENT_USER = config.management_guest_username_default
            constants.CSLE_GUEST.MANAGEMENT_PW = config.management_guest_password_default
            constants.CSLE_GUEST.MANAGEMENT_EMAIL = config.management_guest_email_default
            constants.CSLE_GUEST.MANAGEMENT_ORGANIZATION = config.management_guest_organization_default
            constants.CSLE_GUEST.MANAGEMENT_FIRST_NAME = config.management_guest_first_name_default
            constants.CSLE_GUEST.MANAGEMENT_LAST_NAME = config.management_guest_last_name_default
            constants.AGENT.USER = config.ssh_agent_username
            constants.AGENT.PW = config.ssh_agent_password
            constants.METADATA_STORE.USER = config.metastore_user
            constants.METADATA_STORE.PASSWORD = config.metastore_password
            constants.METADATA_STORE.DBNAME = config.metastore_database_name
            constants.COMMANDS.PGADMIN_USERNAME = config.pgadmin_username
            constants.COMMANDS.PGADMIN_PW = config.pgadmin_password
            constants.COMMANDS.GRAFANA_USERNAME = config.grafana_username
            constants.COMMANDS.GRAFANA_PW = config.grafana_password
            constants.COMMANDS.PGADMIN_PW = config.pgadmin_password
            constants.COMMANDS.NODE_EXPORTER_PORT = config.node_exporter_port
            constants.COMMANDS.GRAFANA_PORT = config.grafana_port
            constants.COMMANDS.MANAGEMENT_SYSTEM_PORT = config.management_system_port
            constants.COMMANDS.CADVISOR_PORT = config.cadvisor_port
            constants.COMMANDS.PGADMIN_PORT = config.pgadmin_port
            constants.COMMANDS.PROMETHEUS_PORT = config.prometheus_port
            constants.COMMANDS.NODE_EXPORTER_PID_FILE = config.node_exporter_pid_file
            constants.COMMANDS.CSLE_MGMT_WEBAPP_PID_FILE = config.csle_mgmt_webapp_pid_file
            constants.COMMANDS.NODE_EXPORTER_LOG_FILE = config.node_exporter_log_file
            constants.COMMANDS.DOCKER_STATS_MANAGER_OUTFILE = config.docker_stats_manager_outfile
            constants.COMMANDS.DOCKER_STATS_MANAGER_PIDFILE = config.docker_stats_manager_pidfile
            constants.COMMANDS.PROMETHEUS_PID_FILE = config.prometheus_pid_file
            constants.COMMANDS.PROMETHEUS_LOG_FILE = config.prometheus_log_file
            constants.COMMANDS.POSTGRESQL_LOG_DIR = config.postgresql_log_dir
            constants.COMMANDS.NGINX_LOG_DIR = config.nginx_log_dir
            constants.COMMANDS.FLASK_LOG_FILE = config.flask_log_file
            constants.LOGGING.DEFAULT_LOG_DIR = config.default_log_dir
            Logger.__call__().get_logger().info(f"Successfully initialized configuration "
                                                f"from configuration file: {config_file_path}")
        except Exception as e:
            Logger.__call__().get_logger().info(f"Failed to read configuration file from: {config_file_path}. "
                                                f"Exception: {str(e)}, {repr(e)}")



from typing import Dict, Any, List
from csle_common.dao.emulation_config.emulation_env_config import EmulationEnvConfig
from csle_base.json_serializable import JSONSerializable


class EmulationExecution(JSONSerializable):
    """
    A DTO representing an execution of an emulation
    """

    def __init__(self, emulation_name: str, timestamp: float, ip_first_octet: int,
                 emulation_env_config: EmulationEnvConfig, physical_servers: List[str]):
        """
        Initializes the DTO

        :param emulation_name: the emulation name
        :param timestamp: the timestamp the execution started
        :param ip_first_octet: the first octet in the ip
        :param emulation_env_config: the emulation env config
        :param id: the id of the execution
        :param physical_servers: list of physical server IPs where the execution is running
        """
        self.emulation_name = emulation_name
        self.timestamp = timestamp
        self.emulation_env_config = emulation_env_config
        self.ip_first_octet = ip_first_octet
        self.physical_servers = physical_servers

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> "EmulationExecution":
        """
        Converts a dict representation to a DTO

        :param d: the dict to convert
        :return: the DTO
        """
        return EmulationExecution(emulation_name=d["emulation_name"], timestamp=d["timestamp"],
                                  ip_first_octet=d["ip_first_octet"],
                                  emulation_env_config=EmulationEnvConfig.from_dict(d["emulation_env_config"]),
                                  physical_servers=d["physical_servers"])

    def to_dict(self) -> Dict[str, Any]:
        """
        Converts the object to a dict representation

        :return: a dict representation of the object
        """
        d: Dict[str, Any] = {}
        d["emulation_name"] = self.emulation_name
        d["timestamp"] = self.timestamp
        d["emulation_env_config"] = self.emulation_env_config.to_dict()
        d["ip_first_octet"] = self.ip_first_octet
        d["physical_servers"] = self.physical_servers
        return d

    def __str__(self) -> str:
        """
        :return: a string representation of the object
        """
        return f"emulation_name:{self.emulation_name}, timestamp: {self.timestamp}, " \
               f"emulation_env_config: {self.emulation_env_config}," \
               f"ip_first_octet:{self.ip_first_octet}, physical_servers: {self.physical_servers}"

    @staticmethod
    def from_json_file(json_file_path: str) -> "EmulationExecution":
        """
        Reads a json file and converts it to a DTO

        :param json_file_path: the json file path
        :return: the converted DTO
        """
        import io
        import json
        with io.open(json_file_path, 'r') as f:
            json_str = f.read()
        return EmulationExecution.from_dict(json.loads(json_str))



from typing import Union, List, Dict, Any
import numpy as np
from csle_common.dao.training.agent_type import AgentType
from csle_common.dao.training.player_type import PlayerType
from csle_common.dao.training.policy import Policy
from csle_common.dao.simulation_config.action import Action
from csle_common.dao.simulation_config.state import State
from csle_common.dao.training.policy_type import PolicyType


class AlphaVectorsPolicy(Policy):
    """
    Object representing a policy based on alpha vectors for a POMDP (Sondik 1971)
    """

    def __init__(self, player_type: PlayerType, actions: List[Action], alpha_vectors: List[Any],
                 transition_tensor: List[Any], reward_tensor: List[Any], states: List[State],
                 agent_type: AgentType, simulation_name: str, avg_R: float) -> None:
        """
        Initializes the policy

        :param actions: list of actions
        :param states: list of states
        :param player_type: the player type
        :param alpha_vectors: the lookup table that defines the policy
        :param value_function: the value function (optional)
        :param simulation_name: the name of the simulation
        :param avg_R: average reward obtained with the policy
        :param transition_tensor: the transition tensor of the POMDP
        :param reward_tensor: the reward tensor of the POMDP
        """
        super(AlphaVectorsPolicy, self).__init__(agent_type=agent_type, player_type=player_type)
        self.actions = actions
        self.alpha_vectors = alpha_vectors
        self.simulation_name = simulation_name
        self.id = -1
        self.avg_R = avg_R
        self.transition_tensor = transition_tensor
        self.reward_tensor = reward_tensor
        self.states = states
        self.policy_type = PolicyType.ALPHA_VECTORS

    def action(self, o: List[Union[int, float]], deterministic: bool = True) -> int:
        """
        Selects the next action

        :param o: the belief
        :param deterministic: boolean flag indicating whether the action selection should be deterministic
        :return: the next action and its probability
        """
        b = o
        max_a_v = -np.inf
        for a in self.actions:
            v_a = 0
            for s in self.states:
                for s_prime in self.states:
                    transition_prob = (b[s.id] * self.reward_tensor[a.id][s.id] *
                                       self.transition_tensor[a.id][s.id][s_prime.id])
                    max_alpha_v = -np.inf
                    for alpha in self.alpha_vectors:
                        v = np.dot(np.array(alpha), np.array(b[0:len(alpha)]))
                        if v > max_alpha_v:
                            max_alpha_v = v
                    v_a += max_alpha_v * transition_prob
            if v_a > max_a_v:
                max_a_v = v_a
                max_a = a
        try:
            return max_a.id
        except Exception:
            return 0

    def probability(self, o: List[Union[int, float]], a: int) -> float:
        """
        Calculates the probability of taking a given action for a given observation

        :param o: the input observation
        :param a: the action
        :return: p(a|o)
        """
        return a == self.action(o=o)

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> "AlphaVectorsPolicy":
        """
        Converts a dict representation to an instance

        :param d: the dict to convert
        :return: the created instance
        """
        dto = AlphaVectorsPolicy(actions=list(map(lambda x: Action.from_dict(x), d["actions"])),
                                 player_type=d["player_type"], agent_type=d["agent_type"],
                                 alpha_vectors=d["alpha_vectors"],
                                 simulation_name=d["simulation_name"], avg_R=d["avg_R"],
                                 transition_tensor=d["transition_tensor"], reward_tensor=d["reward_tensor"],
                                 states=list(map(lambda x: State.from_dict(x), d["states"])))
        if "id" in d:
            dto.id = d["id"]
        return dto

    def to_dict(self) -> Dict[str, Any]:
        """
        :return: A dict representation of the function
        """
        d: Dict[str, Any] = {}
        d["agent_type"] = self.agent_type
        d["player_type"] = self.player_type
        d["actions"] = list(map(lambda x: x.to_dict(), self.actions))
        d["alpha_vectors"] = self.alpha_vectors
        d["simulation_name"] = self.simulation_name
        d["id"] = self.id
        d["avg_R"] = self.avg_R
        d["transition_tensor"] = self.transition_tensor
        d["states"] = list(map(lambda x: x.to_dict(), self.states))
        d["reward_tensor"] = self.reward_tensor
        d["policy_type"] = self.policy_type
        return d

    def stage_policy(self, o: Union[List[Union[int, float]], int, float]) -> List[List[float]]:
        """
        Gets the stage policy, i.e a |S|x|A| policy

        :param o: the latest observation
        :return: the |S|x|A| stage policy
        """
        return self.alpha_vectors

    def __str__(self) -> str:
        """
        :return: a string representation of the policy
        """
        return f"agent_type: {self.agent_type}, player_type: {self.player_type}, " \
               f"actions: {list(map(lambda x: str(x), self.actions))}, alpha_vectors: {self.alpha_vectors}, " \
               f"simulation_name: {self.simulation_name}, id: {self.id}, avg_R: {self.avg_R}," \
               f"transition_tensor: {self.transition_tensor}, states: {self.states}, " \
               f"reward_tensor: {self.reward_tensor}, policy_type: {self.policy_type}"

    @staticmethod
    def from_json_file(json_file_path: str) -> "AlphaVectorsPolicy":
        """
        Reads a json file and converts it to a DTO

        :param json_file_path: the json file path
        :return: the converted DTO
        """
        import io
        import json
        with io.open(json_file_path, 'r') as f:
            json_str = f.read()
        return AlphaVectorsPolicy.from_dict(json.loads(json_str))

    def copy(self) -> "AlphaVectorsPolicy":
        """
        :return: a copy of the DTO
        """
        return AlphaVectorsPolicy.from_dict(self.to_dict())



from typing import Any, Dict
from csle_common.dao.training.policy import Policy
from csle_common.dao.training.multi_threshold_stopping_policy import MultiThresholdStoppingPolicy
from csle_common.dao.training.ppo_policy import PPOPolicy
from csle_base.json_serializable import JSONSerializable


class ExperimentResult(JSONSerializable):
    """
    DTO representing the results of an experiment
    """

    def __init__(self):
        """
        Initializes the DTO
        """
        self.all_metrics: Dict[int, Dict[str, Any]] = {}
        self.policies: Dict[int, Policy] = {}
        self.plot_metrics = []
        self.avg_metrics = {}
        self.std_metrics = {}

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> "ExperimentResult":
        """
        Converts a dict representation to an instance

        :param d: the dict to convert
        :return: the created instance
        """
        obj = ExperimentResult()
        obj.std_metrics = d["std_metrics"]
        obj.avg_metrics = d["avg_metrics"]
        obj.all_metrics = d["all_metrics"]
        obj.plot_metrics = d["plot_metrics"]
        d2 = {}
        for k, v in d["policies"].items():
            try:
                d2[k] = MultiThresholdStoppingPolicy.from_dict(v)
            except Exception:
                pass
            try:
                d2[k] = PPOPolicy.from_dict(v)
            except Exception:
                pass
        obj.policies = d2
        return obj

    def to_dict(self) -> Dict[str, Any]:
        """
        Converts the object to a dict representation

        :return: a dict representation of the object
        """
        d: Dict[str, Any] = {}
        d["all_metrics"] = self.all_metrics
        d2 = {}
        for k, v in self.policies.items():
            d2[k] = v.to_dict()
        d["policies"] = d2
        d["plot_metrics"] = self.plot_metrics
        d["avg_metrics"] = self.avg_metrics
        d["std_metrics"] = self.std_metrics
        return d

    def __str__(self) -> str:
        """
        :return: a string representation of the object
        """
        return f"all_metrics: {self.all_metrics}, policies: {self.policies}, plot_metrics: {self.plot_metrics}, " \
               f"avg_metrics: {self.avg_metrics}, std metrics: {self.std_metrics}"

    @staticmethod
    def from_json_file(json_file_path: str) -> "ExperimentResult":
        """
        Reads a json file and converts it to a DTO

        :param json_file_path: the json file path
        :return: the converted DTO
        """
        import io
        import json
        with io.open(json_file_path, 'r') as f:
            json_str = f.read()
        return ExperimentResult.from_dict(json.loads(json_str))



from typing import Optional, List, Dict, Any, Union
from csle_common.dao.emulation_config.vulnerability_type import VulnType
from csle_common.dao.emulation_config.transport_protocol import TransportProtocol
from csle_common.dao.emulation_config.credential import Credential
from csle_common.util.general_util import GeneralUtil
from csle_base.json_serializable import JSONSerializable


class NodeVulnerabilityConfig(JSONSerializable):
    """
    A DTO object representing a vulnerability of a container in the emulation environment
    """

    def __init__(self, ip: str, vuln_type: VulnType, name: str, port: Union[int, None],
                 protocol: TransportProtocol, credentials: Optional[List[Credential]] = None,
                 cvss: float = 2.0, cve: Optional[str] = None, service: Optional[str] = None, root: bool = False,
                 docker_gw_bridge_ip: str = "", physical_host_ip: str = ""):
        """
        Initializes the DTO

        :param ip: the ip of the node
        :param vuln_type: the vulnerability type
        :param name: the name of the vulnerability
        :param port: the port of the service of the vulnerability
        :param protocol: the protocol of the service of the vulnerability
        :param credentials: the credentials of the vulnerability
        :param cvss: the CVSS of the vulnerability
        :param cve: the CVE of the vulnerability
        :param service: the service of the vulnerability
        :param root: if the vulnerability gives root or not
        :param docker_gw_bridge_ip: IP to reach the container from the host network
        :param physical_host_ip: IP of the physical host where the container is running
        """
        self.ip = ip
        self.vuln_type = vuln_type
        self.name = name
        self.port = port
        self.protocol = protocol
        self.credentials = credentials
        self.cvss = cvss
        self.cve = cve
        self.service = service
        self.root = root
        self.docker_gw_bridge_ip = docker_gw_bridge_ip
        self.physical_host_ip = physical_host_ip

    def to_dict(self) -> Dict[str, Any]:
        """
        Converts the object to a dict representation

        :return: a dict representation of the object
        """
        d: Dict[str, Any] = {}
        d["ip"] = self.ip
        d["vuln_type"] = self.vuln_type
        d["name"] = self.name
        d["port"] = self.port
        d["protocol"] = self.protocol
        d["credentials"] = list(map(lambda x: x.to_dict(), self.credentials)) if self.credentials is not None else []
        d["root"] = self.root
        d["cve"] = self.cve
        d["docker_gw_bridge_ip"] = self.docker_gw_bridge_ip
        d["physical_host_ip"] = self.physical_host_ip
        return d

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> "NodeVulnerabilityConfig":
        """
        Convert a dict representation to a DTO representation

        :return: a dto representation of the object
        """
        dto = NodeVulnerabilityConfig(ip=d["ip"], vuln_type=d["vuln_type"],
                                      name=d["name"], port=d["port"],
                                      protocol=d["protocol"],
                                      credentials=list(map(lambda x: Credential.from_dict(x), d["credentials"])),
                                      root=d["root"], cve=d["cve"],
                                      docker_gw_bridge_ip=d["docker_gw_bridge_ip"],
                                      physical_host_ip=d["physical_host_ip"])
        return dto

    def __str__(self) -> str:
        """
        :return: a string representation of the object
        """
        return f"node_internal_ip:{self.ip}, vuln_type:{self.vuln_type}, name:{self.name}, port:{self.port}, " \
               f"protocol: {self.protocol}, " \
               f"credentials: {list(map(lambda x: str(x), self.credentials)) if self.credentials is not None else []}" \
               f", " \
               f"cvss: {self.cvss}, cve: {self.cve}, service: {self.service}, root:{self.root}, " \
               f"docker_gw_bridge_ip: {self.docker_gw_bridge_ip}, physical_host_ip: {self.physical_host_ip}"

    @staticmethod
    def from_json_file(json_file_path: str) -> "NodeVulnerabilityConfig":
        """
        Reads a json file and converts it to a DTO

        :param json_file_path: the json file path
        :return: the converted DTO
        """
        import io
        import json
        with io.open(json_file_path, 'r') as f:
            json_str = f.read()
        return NodeVulnerabilityConfig.from_dict(json.loads(json_str))

    def copy(self) -> "NodeVulnerabilityConfig":
        """
        :return: a copy of the DTO
        """
        return NodeVulnerabilityConfig.from_dict(self.to_dict())

    def create_execution_config(self, ip_first_octet: int) -> "NodeVulnerabilityConfig":
        """
        Creates a new config for an execution

        :param ip_first_octet: the first octet of the IP of the new execution
        :return: the new config
        """
        config = self.copy()
        config.ip = GeneralUtil.replace_first_octet_of_ip(ip=config.ip, ip_first_octet=ip_first_octet)
        return config



from typing import List, Dict, Any
from csle_common.util.general_util import GeneralUtil
from csle_base.json_serializable import JSONSerializable


class OvsSwitchConfig(JSONSerializable):
    """
    DTO containing the configuration of an OVS Switch
    """

    def __init__(self, container_name: str, ip: str, openflow_protocols: List[str], controller_ip: str,
                 controller_port: int, controller_transport_protocol: str, docker_gw_bridge_ip: str = "",
                 physical_host_ip: str = ""):
        """
        Initializes the DTO

        :param container_name: the name of the switch container
        :param ip: the ip of the switch container
        :param openflow_protocols: the supported open flow protocols
        :param controller_ip: the ip of the SDN controller
        :param controller_port: the port of the SDN controller
        :param controller_transport_protocol: the transport protocol of the controller
        :param docker_gw_bridge_ip: IP to reach the container from the host network
        :param physical_host_ip: IP of the physical host where the container is running
        """
        self.container_name = container_name
        self.ip = ip
        self.openflow_protocols = openflow_protocols
        self.controller_ip = controller_ip
        self.controller_port = controller_port
        self.controller_transport_protocol = controller_transport_protocol
        self.docker_gw_bridge_ip = docker_gw_bridge_ip
        self.physical_host_ip = physical_host_ip

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> "OvsSwitchConfig":
        """
        Converts a dict representation to an instance

        :param d: the dict to convert
        :return: the created instance
        """
        obj = OvsSwitchConfig(container_name=d["container_name"], ip=d["ip"],
                              openflow_protocols=d["openflow_protocols"], controller_ip=d["controller_ip"],
                              controller_port=d["controller_port"],
                              controller_transport_protocol=d["controller_transport_protocol"],
                              docker_gw_bridge_ip=d["docker_gw_bridge_ip"],
                              physical_host_ip=d["physical_host_ip"])
        return obj

    def to_dict(self) -> Dict[str, Any]:
        """
        Converts the object to a dict representation
        
        :return: a dict representation of the object
        """
        d: Dict[str, Any] = {}
        d["ip"] = self.ip
        d["container_name"] = self.container_name
        d["openflow_protocols"] = self.openflow_protocols
        d["controller_ip"] = self.controller_ip
        d["controller_port"] = self.controller_port
        d["controller_transport_protocol"] = self.controller_transport_protocol
        d["docker_gw_bridge_ip"] = self.docker_gw_bridge_ip
        d["physical_host_ip"] = self.physical_host_ip
        return d

    def __str__(self) -> str:
        """
        :return: a string representation of the object
        """
        return f"ip:{self.ip}, container_name: {self.container_name}, openflow_protocols: {self.openflow_protocols}, " \
               f"controller_ip: {self.controller_ip}, controller_port: {self.controller_port}, " \
               f"controller_transport_protocol: {self.controller_transport_protocol}, " \
               f"docker_gw_bridge_ip: {self.docker_gw_bridge_ip}, physical_host_ip: {self.physical_host_ip}"

    @staticmethod
    def from_json_file(json_file_path: str) -> "OvsSwitchConfig":
        """
        Reads a json file and converts it to a DTO

        :param json_file_path: the json file path
        :return: the converted DTO
        """
        import io
        import json
        with io.open(json_file_path, 'r') as f:
            json_str = f.read()
        return OvsSwitchConfig.from_dict(json.loads(json_str))

    def copy(self) -> "OvsSwitchConfig":
        """
        :return: a copy of the DTO
        """
        return OvsSwitchConfig.from_dict(self.to_dict())

    def create_execution_config(self, ip_first_octet: int, physical_servers: List[str]) -> "OvsSwitchConfig":
        """
        Creates a new config for an execution

        :param ip_first_octet: the first octet of the IP of the new execution
        :param physical_servers: the list of physical servers of the execution
        :return: the new config
        """
        config = self.copy()
        config.ip = GeneralUtil.replace_first_octet_of_ip(ip=config.ip, ip_first_octet=ip_first_octet)
        config.controller_ip = GeneralUtil.replace_first_octet_of_ip(ip=config.controller_ip,
                                                                     ip_first_octet=ip_first_octet)
        config.container_name = f"{config.container_name}-{ip_first_octet}"
        config.physical_host_ip = physical_servers[0]  # TODO Update this
        return config



from typing import Dict, Any
from csle_common.dao.training.hparam import HParam
from csle_common.dao.system_identification.system_model_type import SystemModelType
from csle_base.json_serializable import JSONSerializable


class SystemIdentificationConfig(JSONSerializable):
    """
    DTO representing the configuration of a system identification job
    """

    def __init__(self, model_type: SystemModelType, hparams: Dict[str, HParam], output_dir: str,
                 title: str, log_every: int):
        """
        Initializes the DTO

        :param hparams: a dict with the hyperparameters for the system identification
        :param model_type: the type of system model
        :param output_dir: the output directory
        """
        self.hparams = hparams
        self.model_type = model_type
        self.output_dir = output_dir
        self.title = title
        self.log_every = log_every

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> "SystemIdentificationConfig":
        """
        Converts a dict representation to an instance

        :param d: the dict to convert
        :return: the created instance
        """
        h_d = {}
        for k, v in d["hparams"].items():
            h_d[k] = HParam.from_dict(v)
        obj = SystemIdentificationConfig(hparams=h_d, model_type=d["model_type"], output_dir=d["output_dir"],
                                         title=d["title"], log_every=d["log_every"])
        return obj

    def to_dict(self) -> Dict[str, Any]:
        """
        Converts the object to a dict representation
        
        :return: a dict representation of the object
        """
        d: Dict[str, Any] = {}
        d_h = {}
        for k, v in self.hparams.items():
            d_h[k] = v.to_dict()
        d["hparams"] = d_h
        d["model_type"] = self.model_type
        d["output_dir"] = self.output_dir
        d["title"] = self.title
        d["log_every"] = self.log_every
        return d

    def __str__(self) -> str:
        """
        :return: a string representation of the DTO
        """
        return f"hparams: {self.hparams}, model_type: {self.model_type}, output_dir: {self.output_dir}, " \
               f"title: {self.title}, log_every: {self.log_every}"

    @staticmethod
    def from_json_file(json_file_path: str) -> "SystemIdentificationConfig":
        """
        Reads a json file and converts it to a DTO

        :param json_file_path: the json file path
        :return: the converted DTO
        """
        import io
        import json
        with io.open(json_file_path, 'r') as f:
            json_str = f.read()
        return SystemIdentificationConfig.from_dict(json.loads(json_str))



from typing import List, Dict, Any
from csle_base.json_serializable import JSONSerializable


class InitialStateDistributionConfig(JSONSerializable):
    """
    DTO class representing the configuration of the initial state distribution of a simulation
    """

    def __init__(self, initial_state_distribution: List[float]):
        """
        Initializes the DTO

        :param initial_state_distribution: the initial state distribution
        """
        assert round(sum(initial_state_distribution), 3) == 1
        self.initial_state_distribution = initial_state_distribution

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> "InitialStateDistributionConfig":
        """
        Converts a dict representation to an instance

        :param d: the dict to convert
        :return: the created instance
        """
        obj = InitialStateDistributionConfig(
            initial_state_distribution=d["initial_state_distribution"]
        )
        return obj

    def to_dict(self) -> Dict[str, Any]:
        """
        Converts the object to a dict representation
        
        :return: a dict representation of the object
        """
        d = {}
        d["initial_state_distribution"] = self.initial_state_distribution
        return d

    def __str__(self):
        """
        :return: a string representation of the object
        """
        return f"initial_state_distribution: {self.initial_state_distribution}"

    @staticmethod
    def from_json_file(json_file_path: str) -> "InitialStateDistributionConfig":
        """
        Reads a json file and converts it to a DTO

        :param json_file_path: the json file path
        :return: the converted DTO
        """
        import io
        import json
        with io.open(json_file_path, 'r') as f:
            json_str = f.read()
        return InitialStateDistributionConfig.from_dict(json.loads(json_str))



from typing import Optional, List
import csle_common.constants.constants as constants
from csle_common.dao.emulation_action.attacker.emulation_attacker_action import EmulationAttackerAction
from csle_common.dao.emulation_action.attacker.emulation_attacker_action_type import EmulationAttackerActionType
from csle_common.dao.emulation_action.attacker.emulation_attacker_action_id import EmulationAttackerActionId
from csle_common.dao.emulation_action.attacker.emulation_attacker_action_outcome import EmulationAttackerActionOutcome


class EmulationAttackerNMAPActions:
    """
    Class containing Attacker NMAP actions in the emulation
    """

    @staticmethod
    def TCP_SYN_STEALTH_SCAN(index: int, ips: Optional[List[str]] = None) -> EmulationAttackerAction:
        """
        Runs a TCP SYN scan

        :param index: index of the machine to apply the action to
        :param ips: ips of the machines or subnets to apply the action to
        :return: the action
        """
        id = EmulationAttackerActionId.TCP_SYN_STEALTH_SCAN_HOST
        if ips is None:
            ips = []
        if index == -1:
            id = EmulationAttackerActionId.TCP_SYN_STEALTH_SCAN_ALL

        cmd = ["sudo nmap -sS -p- " + constants.NMAP.SPEED_ARGS + " "]
        return EmulationAttackerAction(id=id, name="TCP SYN (Stealth) Scan", cmds=cmd,
                                       type=EmulationAttackerActionType.RECON,
                                       descr="A stealthy and fast TCP SYN scan to detect open TCP ports on the subnet",
                                       ips=ips, index=index,
                                       action_outcome=EmulationAttackerActionOutcome.INFORMATION_GATHERING,
                                       backdoor=False)

    @staticmethod
    def PING_SCAN(index: int, ips: Optional[List[str]] = None) -> EmulationAttackerAction:
        """
        Runs a Ping scan

        :param index: index of the machine to apply the action to
        :param ips: ip of the machine or subnet to apply the action to
        :return: the action
        """
        id = EmulationAttackerActionId.PING_SCAN_HOST
        if ips is None:
            ips = []
        if index == -1:
            id = EmulationAttackerActionId.PING_SCAN_ALL

        cmd = ["sudo nmap -sP " + constants.NMAP.SPEED_ARGS + " "]
        return EmulationAttackerAction(id=id, name="Ping Scan", cmds=cmd,
                                       type=EmulationAttackerActionType.RECON,
                                       descr="A host discovery scan, it is quick because it only checks of hosts "
                                             "are up with Ping, without scanning the ports.",
                                       ips=ips, index=index,
                                       action_outcome=EmulationAttackerActionOutcome.INFORMATION_GATHERING,
                                       backdoor=False)

    @staticmethod
    def UDP_PORT_SCAN(index: int, ips: Optional[List[str]] = None) -> EmulationAttackerAction:
        """
        Runs a UDP port scan

        :param index: index of the machine to apply the action to
        :param ips: ips of the machines or subnets to apply the action to
        :return: the action
        """
        id = EmulationAttackerActionId.UDP_PORT_SCAN_HOST
        if ips is None:
            ips = []
        if index == -1:
            id = EmulationAttackerActionId.UDP_PORT_SCAN_ALL

        cmd = ["sudo nmap -sU -p- " + constants.NMAP.SPEED_ARGS + " "]
        return EmulationAttackerAction(id=id, name="UDP Port Scan", cmds=cmd,
                                       type=EmulationAttackerActionType.RECON,
                                       descr="", index=index,
                                       ips=ips,
                                       action_outcome=EmulationAttackerActionOutcome.INFORMATION_GATHERING,
                                       backdoor=False)

    @staticmethod
    def TCP_CON_NON_STEALTH_SCAN(index: int, ips: Optional[List[str]] = None) -> EmulationAttackerAction:
        """
        Runs a TCP CON (non-stealthy) scan

        :param index: index of the machine to apply the action to
        :param ips: ips of the machines or subnets to apply the action to
        :return: the action
        """
        id = EmulationAttackerActionId.TCP_CON_NON_STEALTH_SCAN_HOST
        if ips is None:
            ips = []

        if index == -1:
            id = EmulationAttackerActionId.TCP_CON_NON_STEALTH_SCAN_ALL

        cmd = ["sudo nmap -sT -p- " + constants.NMAP.SPEED_ARGS + " "]
        return EmulationAttackerAction(
            id=id, name="TCP Connection (Non-Stealth) Scan", cmds=cmd,
            type=EmulationAttackerActionType.RECON, index=index,
            descr="A non-stealthy and fast TCP SYN scan to detect open TCP ports on the subnet",
            ips=ips, action_outcome=EmulationAttackerActionOutcome.INFORMATION_GATHERING,
            backdoor=False)

    @staticmethod
    def TCP_FIN_SCAN(index: int, ips: Optional[List[str]] = None) -> EmulationAttackerAction:
        """
        Runs a TCP FIN scan

        :param index: index of the machine to apply the action to
        :param ips: ips of the machines or subnets to apply the action to
        :return: the action
        """
        id = EmulationAttackerActionId.TCP_FIN_SCAN_HOST
        if ips is None:
            ips = []
        if index == -1:
            id = EmulationAttackerActionId.TCP_FIN_SCAN_ALL

        cmd = ["sudo nmap -sF -p- " + constants.NMAP.SPEED_ARGS + " "]
        return EmulationAttackerAction(
            id=id, name="FIN Scan", cmds=cmd,
            type=EmulationAttackerActionType.RECON, index=index,
            descr="A special type of TCP port scan using FIN, can avoid IDS and firewalls that block SYN scans",
            ips=ips, action_outcome=EmulationAttackerActionOutcome.INFORMATION_GATHERING,
            backdoor=False)

    @staticmethod
    def TCP_NULL_SCAN(index: int, ips: Optional[List[str]] = None) -> EmulationAttackerAction:
        """
        Runs a TCP Null scan

        :param index: index of the machine to apply the action to
        :param ips: ips of the machines or subnets to apply the action to
        :return: the action
        """
        id = EmulationAttackerActionId.TCP_NULL_SCAN_HOST
        if ips is None:
            ips = []
        if index == -1:
            id = EmulationAttackerActionId.TCP_NULL_SCAN_ALL

        cmd = ["sudo nmap -sN -p- " + constants.NMAP.SPEED_ARGS + " "]
        return EmulationAttackerAction(
            id=id, name="Null Scan", cmds=cmd, index=index, type=EmulationAttackerActionType.RECON,
            descr="A special type of TCP port scan using Null, can avoid IDS and firewalls that block SYN scans",
            ips=ips, action_outcome=EmulationAttackerActionOutcome.INFORMATION_GATHERING, backdoor=False)

    @staticmethod
    def TCP_XMAS_TREE_SCAN(index: int, ips: Optional[List[str]] = None) -> EmulationAttackerAction:
        """
        Runs a TCP XMAS TREE scan

        :param index: index of the machine to apply the action to
        :param ips: ips of the machines or subnets to apply the action to
        :return: the action
        """
        id = EmulationAttackerActionId.TCP_XMAS_TREE_SCAN_HOST
        if ips is None:
            ips = []
        if index == -1:
            id = EmulationAttackerActionId.TCP_XMAS_TREE_SCAN_ALL

        cmd = ["sudo nmap -sX -p- " + constants.NMAP.SPEED_ARGS + " "]
        return EmulationAttackerAction(id=id, name="Xmas Tree Scan",
                                       cmds=cmd, type=EmulationAttackerActionType.RECON, index=index,
                                       descr="A special type of TCP port scan using XMas Tree, "
                                             "can avoid IDS and firewalls that block SYN scans",
                                       ips=ips, action_outcome=EmulationAttackerActionOutcome.INFORMATION_GATHERING,
                                       backdoor=False)

    @staticmethod
    def OS_DETECTION_SCAN(index: int, ips: Optional[List[str]] = None) -> EmulationAttackerAction:
        """
        Runs a OS detection scan

        :param index: index of the machine to apply the action to
        :param ips: ips of the machines or subnets to apply the action to
        :return: the action
        """
        id = EmulationAttackerActionId.OS_DETECTION_SCAN_HOST
        if ips is None:
            ips = []
        if index == -1:
            id = EmulationAttackerActionId.OS_DETECTION_SCAN_ALL

        cmd = ["sudo nmap -O --osscan-guess --max-os-tries 1 " + constants.NMAP.SPEED_ARGS + " "]
        return EmulationAttackerAction(id=id, name="OS detection scan",
                                       cmds=cmd, type=EmulationAttackerActionType.RECON,
                                       descr="OS detection/guess scan", index=index,
                                       ips=ips, action_outcome=EmulationAttackerActionOutcome.INFORMATION_GATHERING,
                                       backdoor=False)

    @staticmethod
    def VULSCAN(index: int, ips: Optional[List[str]] = None) -> EmulationAttackerAction:
        """
        Runs a vulnerability scan using the VULSCAN script

        :param index: index of the machine to apply the action to
        :param ips: ips of the machines or subnets to apply the action to
        :return: the action
        """
        id = EmulationAttackerActionId.VULSCAN_HOST
        if ips is None:
            ips = []
        if index == -1:
            id = EmulationAttackerActionId.VULSCAN_ALL

        cmd = ["sudo nmap -sV --script=vulscan/vulscan.nse " + constants.NMAP.SPEED_ARGS + " "]
        return EmulationAttackerAction(id=id, name="vulscan.nse vulnerability scanner",
                                       cmds=cmd, type=EmulationAttackerActionType.RECON, index=index,
                                       descr="Uses a vulcan.nse script to turn NMAP into a vulnerability scanner",
                                       ips=ips, action_outcome=EmulationAttackerActionOutcome.INFORMATION_GATHERING,
                                       backdoor=False)

    @staticmethod
    def NMAP_VULNERS(index: int, ips: Optional[List[str]] = None) -> EmulationAttackerAction:
        """
        Runs a vulnerability scan using the Vulners script

        :param index: index of the machine to apply the action to
        :param ips: ip of the machine or subnet to apply the action to
        :return: the action
        """
        id = EmulationAttackerActionId.NMAP_VULNERS_HOST
        if ips is None:
            ips = []
        if index == -1:
            id = EmulationAttackerActionId.NMAP_VULNERS_ALL

        cmd = ["sudo nmap -sV --script vulners.nse " + constants.NMAP.SPEED_ARGS + " "]
        return EmulationAttackerAction(id=id, name="nmap_vulners vulnerability scanner",
                                       cmds=cmd, type=EmulationAttackerActionType.RECON, index=index,
                                       descr="Uses vulners.nse script to turn NMAP into a vulnerability scanner",
                                       ips=ips, action_outcome=EmulationAttackerActionOutcome.INFORMATION_GATHERING,
                                       backdoor=False)

    @staticmethod
    def TELNET_SAME_USER_PASS_DICTIONARY(index: int, ips: Optional[List[str]] = None) -> EmulationAttackerAction:
        """
        Runs a dictionary attack trying combinations with same user+pw against telnet

        :param index: index of the machine to apply the action to
        :param ips: ips of the machines or subnets to apply the action to
        :return: the action
        """
        telnet_args = constants.NMAP.TELNET_BRUTE_HOST
        id = EmulationAttackerActionId.TELNET_SAME_USER_PASS_DICTIONARY_HOST
        if ips is None:
            ips = []
        if index == -1:
            id = EmulationAttackerActionId.TELNET_SAME_USER_PASS_DICTIONARY_ALL
            telnet_args = constants.NMAP.TELNET_BRUTE_SUBNET

        cmd = ["sudo nmap " + telnet_args + " " + constants.NMAP.SPEED_ARGS + " "]
        return EmulationAttackerAction(
            id=id, name="Telnet dictionary attack for username=pw",
            cmds=cmd, type=EmulationAttackerActionType.EXPLOIT, index=index,
            descr="A dictionary attack that tries common passwords and usernames "
                  "for Telnet where username=password",
            ips=ips, action_outcome=EmulationAttackerActionOutcome.SHELL_ACCESS,
            vulnerability=constants.EXPLOIT_VULNERABILITES.TELNET_DICTS_SAME_USER_PASS,
            backdoor=False)

    @staticmethod
    def SSH_SAME_USER_PASS_DICTIONARY(index: int, ips: Optional[List[str]] = None) -> EmulationAttackerAction:
        """
        Runs a dictionary attack trying combinations with same user+pw against ssh

        :param index: index of the machine to apply the action to
        :param ips: ip of the machine or subnet to apply the action to
        :return: the action
        """
        ssh_args = constants.NMAP.SSH_BRUTE_HOST
        id = EmulationAttackerActionId.SSH_SAME_USER_PASS_DICTIONARY_HOST
        if ips is None:
            ips = []
        if index == -1:
            id = EmulationAttackerActionId.SSH_SAME_USER_PASS_DICTIONARY_ALL
            ssh_args = constants.NMAP.SSH_BRUTE_SUBNET

        cmd = ["sudo nmap " + ssh_args + " " + constants.NMAP.SPEED_ARGS + " "]
        return EmulationAttackerAction(id=id, name="SSH dictionary attack for username=pw",
                                       cmds=cmd, type=EmulationAttackerActionType.EXPLOIT, index=index,
                                       descr="A dictionary attack that tries common passwords and usernames "
                                             "for SSH where username=password",
                                       ips=ips, action_outcome=EmulationAttackerActionOutcome.SHELL_ACCESS,
                                       vulnerability=constants.EXPLOIT_VULNERABILITES.SSH_DICT_SAME_USER_PASS,
                                       backdoor=False)

    @staticmethod
    def FTP_SAME_USER_PASS_DICTIONARY(index: int, ips: Optional[List[str]] = None) -> EmulationAttackerAction:
        """
        Runs a dictionary attack trying combinations with same user+pw against ftp

        :param index: index of the machine to apply the action to
        :param ips: ip of the machine or subnet to apply the action to
        :return: the action
        """
        ftp_args = constants.NMAP.FTP_BRUTE_HOST
        id = EmulationAttackerActionId.FTP_SAME_USER_PASS_DICTIONARY_HOST
        if ips is None:
            ips = []
        if index == -1:
            id = EmulationAttackerActionId.FTP_SAME_USER_PASS_DICTIONARY_ALL
            ftp_args = constants.NMAP.FTP_BRUTE_SUBNET

        cmd = ["sudo nmap " + ftp_args + " " + constants.NMAP.SPEED_ARGS + " "]
        return EmulationAttackerAction(id=id, name="FTP dictionary attack for username=pw",
                                       cmds=cmd, type=EmulationAttackerActionType.EXPLOIT,
                                       descr="A dictionary attack that tries common passwords and usernames "
                                             "for FTP where username=password", index=index,
                                       ips=ips, action_outcome=EmulationAttackerActionOutcome.SHELL_ACCESS,
                                       vulnerability=constants.EXPLOIT_VULNERABILITES.FTP_DICT_SAME_USER_PASS,
                                       backdoor=False)

    @staticmethod
    def CASSANDRA_SAME_USER_PASS_DICTIONARY(index: int, ips: Optional[List[str]] = None) -> EmulationAttackerAction:
        """
        Runs a dictionary attack trying combinations with same user+pw against cassandra

        :param index: index of the machine to apply the action to
        :param ips: ips of the machines or subnets to apply the action to
        :return: the action
        """
        cassandra_args = constants.NMAP.CASSANDRA_BRUTE_HOST
        id = EmulationAttackerActionId.CASSANDRA_SAME_USER_PASS_DICTIONARY_HOST
        if ips is None:
            ips = []
        if index == -1:
            id = EmulationAttackerActionId.CASSANDRA_SAME_USER_PASS_DICTIONARY_ALL
            cassandra_args = constants.NMAP.CASSANDRA_BRUTE_SUBNET

        cmd = ["sudo nmap " + cassandra_args + " " + constants.NMAP.SPEED_ARGS + " "]
        return EmulationAttackerAction(id=id, name="Cassandra dictionary attack for username=pw",
                                       cmds=cmd, type=EmulationAttackerActionType.EXPLOIT,
                                       descr="A dictionary attack that tries common passwords and usernames "
                                             "for Cassandra where username=password", index=index,
                                       ips=ips, action_outcome=EmulationAttackerActionOutcome.SHELL_ACCESS,
                                       vulnerability=constants.EXPLOIT_VULNERABILITES.CASSANDRA_DICTS_SAME_USER_PASS,
                                       backdoor=False)

    @staticmethod
    def IRC_SAME_USER_PASS_DICTIONARY(index: int, ips: Optional[List[str]] = None) -> EmulationAttackerAction:
        """
        Runs a dictionary attack trying combinations with same user+pw against irc

        :param index: index of the machine to apply the action to
        :param ips: ips of the machines or subnets to apply the action to
        :return: the action
        """
        irc_args = constants.NMAP.IRC_BRUTE_HOST
        id = EmulationAttackerActionId.IRC_SAME_USER_PASS_DICTIONARY_HOST
        if ips is None:
            ips = []
        if index == -1:
            id = EmulationAttackerActionId.IRC_SAME_USER_PASS_DICTIONARY_ALL
            irc_args = constants.NMAP.IRC_BRUTE_SUBNET

        cmd = ["sudo nmap " + irc_args + " " + constants.NMAP.SPEED_ARGS + " "]
        return EmulationAttackerAction(id=id, name="IRC dictionary attack for username=pw",
                                       cmds=cmd, type=EmulationAttackerActionType.EXPLOIT,
                                       descr="A dictionary attack that tries common passwords and usernames "
                                             "for IRC where username=password", index=index,
                                       ips=ips, action_outcome=EmulationAttackerActionOutcome.SHELL_ACCESS,
                                       vulnerability=constants.EXPLOIT_VULNERABILITES.IRC_DICTS_SAME_USER_PASS,
                                       backdoor=False)

    @staticmethod
    def MONGO_SAME_USER_PASS_DICTIONARY(index: int, ips: Optional[List[str]] = None) -> EmulationAttackerAction:
        """
        Runs a dictionary attack trying combinations with same user+pw against mongo

        :param index: index of the machine to apply the action to
        :param ips: ips of the machines or subnets to apply the action to
        :return: the action
        """
        mongo_args = constants.NMAP.MONGO_BRUTE_HOST
        id = EmulationAttackerActionId.MONGO_SAME_USER_PASS_DICTIONARY_HOST
        if ips is None:
            ips = []
        if index == -1:
            id = EmulationAttackerActionId.MONGO_SAME_USER_PASS_DICTIONARY_ALL
            mongo_args = constants.NMAP.MONGO_BRUTE_SUBNET

        cmd = ["sudo nmap " + mongo_args + " " + constants.NMAP.SPEED_ARGS + " "]
        return EmulationAttackerAction(id=id, name="MongoDB dictionary attack for username=pw",
                                       cmds=cmd, type=EmulationAttackerActionType.EXPLOIT, index=index,
                                       descr="A dictionary attack that tries common passwords and usernames "
                                             "for MongoDB where username=password",
                                       ips=ips, action_outcome=EmulationAttackerActionOutcome.SHELL_ACCESS,
                                       vulnerability=constants.EXPLOIT_VULNERABILITES.MONGO_DICTS_SAME_USER_PASS,
                                       backdoor=False)

    @staticmethod
    def MYSQL_SAME_USER_PASS_DICTIONARY(index: int, ips: Optional[List[str]] = None) -> EmulationAttackerAction:
        """
        Runs a dictionary attack trying combinations with same user+pw against mysql

        :param index: index of the machine to apply the action to
        :param ips: ip of the machines or subnets to apply the action to
        :return: the action
        """
        mysql_args = constants.NMAP.MYSQL_BRUTE_HOST
        id = EmulationAttackerActionId.MYSQL_SAME_USER_PASS_DICTIONARY_HOST
        if ips is None:
            ips = []
        if index == -1:
            mysql_args = constants.NMAP.MYSQL_BRUTE_SUBNET
            id = EmulationAttackerActionId.MYSQL_SAME_USER_PASS_DICTIONARY_ALL

        cmd = ["sudo nmap " + mysql_args + " " + constants.NMAP.SPEED_ARGS + " "]
        return EmulationAttackerAction(id=id, name="MySQL dictionary attack for username=pw",
                                       cmds=cmd, type=EmulationAttackerActionType.EXPLOIT, index=index,
                                       descr="A dictionary attack that tries common passwords and usernames "
                                             "for MySQL where username=password",
                                       ips=ips, action_outcome=EmulationAttackerActionOutcome.SHELL_ACCESS,
                                       vulnerability=constants.EXPLOIT_VULNERABILITES.MYSQL_DICTS_SAME_USER_PASS,
                                       backdoor=False)

    @staticmethod
    def SMTP_SAME_USER_PASS_DICTIONARY(index: int, ips: Optional[List[str]] = None) -> EmulationAttackerAction:
        """
        Runs a dictionary attack trying combinations with same user+pw against smtp

        :param index: index of the machine to apply the action to
        :param ips: ips of the machines or subnets to apply the action to
        :return: the action
        """
        smtp_args = constants.NMAP.SMTP_BRUTE_HOST
        id = EmulationAttackerActionId.SMTP_SAME_USER_PASS_DICTIONARY_HOST
        if ips is None:
            ips = []
        if index == -1:
            smtp_args = constants.NMAP.SMTP_BRUTE_SUBNET
            id = EmulationAttackerActionId.SMTP_SAME_USER_PASS_DICTIONARY_ALL

        cmd = ["sudo nmap " + smtp_args + " " + constants.NMAP.SPEED_ARGS + " "]
        return EmulationAttackerAction(id=id, name="SMTP dictionary attack for username=pw",
                                       cmds=cmd, type=EmulationAttackerActionType.EXPLOIT, index=index,
                                       descr="A dictionary attack that tries common passwords and usernames "
                                             "for SMTP where username=password",
                                       ips=ips, action_outcome=EmulationAttackerActionOutcome.SHELL_ACCESS,
                                       vulnerability=constants.EXPLOIT_VULNERABILITES.SMTP_DICTS_SAME_USER_PASS,
                                       backdoor=False)

    @staticmethod
    def POSTGRES_SAME_USER_PASS_DICTIONARY(index: int, ips: Optional[List[str]] = None) -> EmulationAttackerAction:
        """
        Runs a dictionary attack trying combinations with same user+pw against postgres

        :param index: index of the machine to apply the action to
        :param ips: ips of the machines or subnets to apply the action to
        :return: the action
        """
        postgres_args = constants.NMAP.POSTGRES_BRUTE_HOST
        id = EmulationAttackerActionId.POSTGRES_SAME_USER_PASS_DICTIONARY_HOST
        if ips is None:
            ips = []
        if index == -1:
            postgres_args = constants.NMAP.POSTGRES_BRUTE_SUBNET
            id = EmulationAttackerActionId.POSTGRES_SAME_USER_PASS_DICTIONARY_ALL

        cmd = ["sudo nmap " + postgres_args + " " + constants.NMAP.SPEED_ARGS + " "]
        return EmulationAttackerAction(id=id, name="Postgres dictionary attack for username=pw",
                                       cmds=cmd, type=EmulationAttackerActionType.EXPLOIT, index=index,
                                       descr="A dictionary attack that tries common passwords and usernames "
                                             "for Postgres where username=password",
                                       ips=ips, action_outcome=EmulationAttackerActionOutcome.SHELL_ACCESS,
                                       vulnerability=constants.EXPLOIT_VULNERABILITES.POSTGRES_DICTS_SAME_USER_PASS,
                                       backdoor=False)

    @staticmethod
    def FIREWALK(index: int, ips: Optional[List[str]] = None) -> EmulationAttackerAction:
        """
        Runs a firewalk scan to try to identify and bypass firewalls

        :param index: index of the machine to apply the action to
        :param ips: ips of the machines or subnets to apply the action to
        :return: the action
        """
        firewalk_args = constants.NMAP.FIREWALK_HOST
        id = EmulationAttackerActionId.FIREWALK_HOST
        if ips is None:
            ips = []
        if index == -1:
            firewalk_args = constants.NMAP.FIREWALK_HOST
            id = EmulationAttackerActionId.FIREWALK_ALL

        cmd = ["sudo nmap " + firewalk_args + " " + constants.NMAP.SPEED_ARGS + " "]
        return EmulationAttackerAction(id=id, name="Firewalk scan",
                                       cmds=cmd, type=EmulationAttackerActionType.RECON, index=index,
                                       descr="Tries to discover firewall rules using an IP TTL expiration technique "
                                             "known as firewalking.",
                                       ips=ips, action_outcome=EmulationAttackerActionOutcome.INFORMATION_GATHERING,
                                       backdoor=False)

    @staticmethod
    def HTTP_ENUM(index: int, ips: Optional[List[str]] = None) -> EmulationAttackerAction:
        """
        Runs a HTTP enumeration scan

        :param index: index of the machine to apply the action to
        :param ips: ips of the machines or subnets to apply the action to
        :return: the action
        """
        http_enum_args = constants.NMAP.HTTP_ENUM
        id = EmulationAttackerActionId.HTTP_ENUM_HOST
        if ips is None:
            ips = []
        if index == -1:
            http_enum_args = constants.NMAP.HTTP_ENUM
            id = EmulationAttackerActionId.HTTP_ENUM_ALL

        cmd = ["sudo nmap " + http_enum_args + " " + constants.NMAP.SPEED_ARGS + " "]
        return EmulationAttackerAction(id=id, name="HTTP Enum",
                                       cmds=cmd, type=EmulationAttackerActionType.RECON, index=index,
                                       descr="Enumerates directories used by popular web applications and servers.",
                                       ips=ips, action_outcome=EmulationAttackerActionOutcome.INFORMATION_GATHERING,
                                       backdoor=False)

    @staticmethod
    def HTTP_GREP(index: int, ips: Optional[List[str]] = None) -> EmulationAttackerAction:
        """
        Runs a HTTP grep scan

        :param index: index of the machine to apply the action to
        :param ips: ips of the machines or subnets to apply the action to
        :return: the action
        """
        http_grep_args = constants.NMAP.HTTP_GREP
        id = EmulationAttackerActionId.HTTP_GREP_HOST

        if ips is None:
            ips = []
        if index == -1:
            http_grep_args = constants.NMAP.HTTP_GREP
            id = EmulationAttackerActionId.HTTP_GREP_ALL

        cmd = ["sudo nmap " + http_grep_args + " " + constants.NMAP.SPEED_ARGS + " "]
        return EmulationAttackerAction(id=id, name="HTTP Grep",
                                       cmds=cmd, type=EmulationAttackerActionType.RECON, index=index,
                                       descr="Spiders a website and attempts to match all pages and urls "
                                             "to find ips and emails.",
                                       ips=ips, action_outcome=EmulationAttackerActionOutcome.INFORMATION_GATHERING,
                                       backdoor=False)

    @staticmethod
    def FINGER(index: int, ips: Optional[List[str]] = None) -> EmulationAttackerAction:
        """
        Runs a fingerprint scan

        :param index: index of the machine to apply the action to
        :param ips: ips of the machines or subnets to apply the action to
        :return: the action
        """
        finger_args = constants.NMAP.FINGER
        id = EmulationAttackerActionId.FINGER_HOST

        if ips is None:
            ips = []
        if index == -1:
            finger_args = constants.NMAP.FINGER
            id = EmulationAttackerActionId.FINGER_ALL

        cmd = ["sudo nmap " + finger_args + " " + constants.NMAP.SPEED_ARGS + " "]
        return EmulationAttackerAction(id=id, name="Finger",
                                       cmds=cmd, type=EmulationAttackerActionType.RECON, index=index,
                                       descr="Attempts to retrieve a list of usernames using the finger service.",
                                       ips=ips, action_outcome=EmulationAttackerActionOutcome.INFORMATION_GATHERING,
                                       backdoor=False)



from typing import Dict, Any, Union
from csle_base.json_serializable import JSONSerializable


class ManagementUser(JSONSerializable):
    """
    DTO representing a management user
    """

    def __init__(self, username: str, password: str, email: str, first_name: str, last_name: str,
                 organization: str, admin: bool, salt: str) -> None:
        """
        Initializes the DTO

        :param username: the username of the user
        :param password: the password of the user
        :param admin: boolean flag whether the user is an admin or not
        :param email: the email of the user
        :param first_name: the first name of the user
        :param last_name: the last name of the user
        :param organization: the organization of the user
        :param salt: the password salt of the user
        """
        self.username = username
        self.password = password
        self.admin = admin
        self.salt = salt
        self.email = email
        self.first_name = first_name
        self.last_name = last_name
        self.organization = organization
        self.id = -1

    def to_dict(self) -> Dict[str, Union[str, bool, int]]:
        """
        Converts the object to a dict representation
        
        :return: a dict representation of the object
        """
        d: Dict[str, Union[str, bool, int]] = {}
        d["username"] = self.username
        d["password"] = self.password
        d["salt"] = self.salt
        d["admin"] = self.admin
        d["email"] = self.email
        d["first_name"] = self.first_name
        d["last_name"] = self.last_name
        d["organization"] = self.organization
        d["id"] = self.id
        return d

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> "ManagementUser":
        """
        Converts a dict representation of the DTO into an instance

        :param d: the dict to convert
        :return: the created instance
        """
        obj = ManagementUser(
            username=d["username"], password=d["password"], admin=d["admin"], salt=d["salt"],
            first_name=d["first_name"], last_name=d["last_name"], organization=d["organization"],
            email=d["email"])
        if "id" in d:
            obj.id = d["id"]
        return obj

    def __str__(self) -> str:
        """
        :return: a string representation of the object
        """
        return f"username: {self.username}, password: {self.password}, admin: {self.admin}, id: {self.id}, " \
               f"salt: {self.salt}, email: {self.email}, first_name: {self.first_name}, last_name: {self.last_name}," \
               f" organization: {self.organization}"

    @staticmethod
    def from_json_file(json_file_path: str) -> "ManagementUser":
        """
        Reads a json file and converts it to a DTO

        :param json_file_path: the json file path
        :return: the converted DTO
        """
        import io
        import json
        with io.open(json_file_path, 'r') as f:
            json_str = f.read()
        return ManagementUser.from_dict(json.loads(json_str))
